# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""EnviroAtlas High-Resolution Land Cover datasets."""

import os
import sys
from collections.abc import Callable, Sequence
from typing import Any, ClassVar, cast

import fiona
import matplotlib.pyplot as plt
import numpy as np
import pyproj
import rasterio
import rasterio.mask
import shapely.geometry
import shapely.ops
import torch
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import GeoDataset
from .utils import BoundingBox, Path, download_url, extract_archive


class EnviroAtlas(GeoDataset):
    """EnviroAtlas dataset covering four cities with prior and weak input data layers.

    The `EnviroAtlas
    <https://doi.org/10.5281/zenodo.5778192>`__ dataset contains NAIP aerial imagery,
    NLCD land cover labels, OpenStreetMap roads, water, waterways, and waterbodies,
    Microsoft building footprint labels, high-resolution land cover labels from the
    EPA EnviroAtlas dataset, and high-resolution land cover prior layers.

    This dataset was organized to accompany the 2022 paper, `"Resolving label
    uncertainty with implicit generative models"
    <https://openreview.net/forum?id=AEa_UepnMDX>`_. More details can be found at
    https://github.com/estherrolf/implicit-posterior.

    If you use this dataset in your research, please cite the following paper:

    * https://openreview.net/forum?id=AEa_UepnMDX

    .. versionadded:: 0.3
    """

    url = 'https://zenodo.org/records/5778193/files/enviroatlas_lotp.zip?download=1'
    filename = 'enviroatlas_lotp.zip'
    md5 = 'bfe601be21c7c001315fc6154be8ef14'

    crs = CRS.from_epsg(3857)
    res = 1

    valid_prior_layers = ('prior', 'prior_no_osm_no_buildings')

    valid_layers = (
        'naip',
        'nlcd',
        'roads',
        'water',
        'waterways',
        'waterbodies',
        'buildings',
        'lc',
        *valid_prior_layers,
    )

    cities = (
        'pittsburgh_pa-2010_1m',
        'durham_nc-2012_1m',
        'austin_tx-2012_1m',
        'phoenix_az-2010_1m',
    )
    splits = (
        [f'{state}-train' for state in cities[:1]]
        + [f'{state}-val' for state in cities[:1]]
        + [f'{state}-test' for state in cities]
        + [f'{state}-val5' for state in cities]
    )

    # these are used to check the integrity of the dataset
    _files = (
        'austin_tx-2012_1m-test_tiles-debuffered',
        'austin_tx-2012_1m-val5_tiles-debuffered',
        'durham_nc-2012_1m-test_tiles-debuffered',
        'durham_nc-2012_1m-val5_tiles-debuffered',
        'phoenix_az-2010_1m-test_tiles-debuffered',
        'phoenix_az-2010_1m-val5_tiles-debuffered',
        'pittsburgh_pa-2010_1m-test_tiles-debuffered',
        'pittsburgh_pa-2010_1m-train_tiles-debuffered',
        'pittsburgh_pa-2010_1m-val5_tiles-debuffered',
        'pittsburgh_pa-2010_1m-val_tiles-debuffered',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_a_naip.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_b_nlcd.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_c_roads.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_d1_waterways.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_d2_waterbodies.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_d_water.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_e_buildings.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_h_highres_labels.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_prior_from_cooccurrences_101_31.tif',
        'austin_tx-2012_1m-test_tiles-debuffered/3009726_sw_prior_from_cooccurrences_101_31_no_osm_no_buildings.tif',
        'spatial_index.geojson',
    )

    p_src_crs = pyproj.CRS('epsg:3857')
    p_transformers: ClassVar[dict[str, CRS]] = {
        'epsg:26917': pyproj.Transformer.from_crs(
            p_src_crs, pyproj.CRS('epsg:26917'), always_xy=True
        ).transform,
        'epsg:26918': pyproj.Transformer.from_crs(
            p_src_crs, pyproj.CRS('epsg:26918'), always_xy=True
        ).transform,
        'epsg:26914': pyproj.Transformer.from_crs(
            p_src_crs, pyproj.CRS('epsg:26914'), always_xy=True
        ).transform,
        'epsg:26912': pyproj.Transformer.from_crs(
            p_src_crs, pyproj.CRS('epsg:26912'), always_xy=True
        ).transform,
    }

    # used to convert the 10 high-res classes labeled as [0, 10, 20, 30, 40, 52, 70, 80,
    # 82, 91, 92] to sequential labels [0, ..., 10]
    raw_enviroatlas_to_idx_map: 'np.typing.NDArray[np.uint8]' = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            2,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            4,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            5,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            6,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            7,
            0,
            8,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            9,
            10,
        ],
        dtype=np.uint8,
    )

    highres_classes = (
        'Unclassified',
        'Water',
        'Impervious Surface',
        'Soil and Barren',
        'Trees and Forest',
        'Shrubs',
        'Grass and Herbaceous',
        'Agriculture',
        'Orchards',
        'Woody Wetlands',
        'Emergent Wetlands',
    )
    highres_cmap = ListedColormap(
        [
            [1.00000000, 1.00000000, 1.00000000],
            [0.00000000, 0.77254902, 1.00000000],
            [0.61176471, 0.61176471, 0.61176471],
            [1.00000000, 0.66666667, 0.00000000],
            [0.14901961, 0.45098039, 0.00000000],
            [0.80000000, 0.72156863, 0.47450980],
            [0.63921569, 1.00000000, 0.45098039],
            [0.86274510, 0.85098039, 0.22352941],
            [0.67058824, 0.42352941, 0.15686275],
            [0.72156863, 0.85098039, 0.92156863],
            [0.42352941, 0.62352941, 0.72156863],
        ]
    )

    def __init__(
        self,
        root: Path = 'data',
        splits: Sequence[str] = ['pittsburgh_pa-2010_1m-train'],
        layers: Sequence[str] = ['naip', 'prior'],
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        prior_as_input: bool = False,
        cache: bool = True,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            root: root directory where dataset can be found
            splits: a list of strings in the format "{state}-{train,val,test}"
                indicating the subset of data to use, for example "ny-train"
            layers: a list containing a subset of ``valid_layers`` indicating which
                layers to load
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            prior_as_input: bool describing whether the prior is used as an input (True)
                or as supervision (False)
            cache: if True, cache file handle to speed up repeated sampling
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: if ``splits`` or ``layers`` are not valid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        for split in splits:
            assert split in self.splits
        assert all([layer in self.valid_layers for layer in layers])
        self.root = root
        self.layers = layers
        self.cache = cache
        self.download = download
        self.checksum = checksum
        self.prior_as_input = prior_as_input

        self._verify()

        super().__init__(transforms)

        # Add all tiles into the index in epsg:3857 based on the included geojson
        mint: float = 0
        maxt: float = sys.maxsize
        with fiona.open(
            os.path.join(root, 'enviroatlas_lotp', 'spatial_index.geojson'), 'r'
        ) as f:
            for i, row in enumerate(f):
                if row['properties']['split'] in splits:
                    box = shapely.geometry.shape(row['geometry'])
                    minx, miny, maxx, maxy = box.bounds
                    coords = (minx, maxx, miny, maxy, mint, maxt)

                    self.index.insert(
                        i,
                        coords,
                        {
                            'naip': row['properties']['naip'],
                            'nlcd': row['properties']['nlcd'],
                            'roads': row['properties']['roads'],
                            'water': row['properties']['water'],
                            'waterways': row['properties']['waterways'],
                            'waterbodies': row['properties']['waterbodies'],
                            'buildings': row['properties']['buildings'],
                            'lc': row['properties']['lc'],
                            'prior_no_osm_no_buildings': row['properties'][
                                'naip'
                            ].replace(
                                'a_naip',
                                'prior_from_cooccurrences_101_31_no_osm_no_buildings',
                            ),
                            'prior': row['properties']['naip'].replace(
                                'a_naip', 'prior_from_cooccurrences_101_31'
                            ),
                        },
                    )

    def __getitem__(self, query: BoundingBox) -> dict[str, Any]:
        """Retrieve image/mask and metadata indexed by query.

        Args:
            query: (minx, maxx, miny, maxy, mint, maxt) coordinates to index

        Returns:
            sample of image/mask and metadata at that index

        Raises:
            IndexError: if query is not found in the index
        """
        hits = self.index.intersection(tuple(query), objects=True)
        filepaths = cast(list[dict[str, str]], [hit.object for hit in hits])

        sample = {'image': [], 'mask': [], 'crs': self.crs, 'bounds': query}

        if len(filepaths) == 0:
            raise IndexError(
                f'query: {query} not found in index with bounds: {self.bounds}'
            )
        elif len(filepaths) == 1:
            filenames = filepaths[0]
            query_geom_transformed = None  # is set by the first layer

            minx, maxx, miny, maxy, mint, maxt = query
            query_box = shapely.geometry.box(minx, miny, maxx, maxy)

            for layer in self.layers:
                fn = filenames[layer]

                with rasterio.open(
                    os.path.join(self.root, 'enviroatlas_lotp', fn)
                ) as f:
                    dst_crs = f.crs.to_string().lower()

                    if query_geom_transformed is None:
                        query_box_transformed = shapely.ops.transform(
                            self.p_transformers[dst_crs], query_box
                        ).envelope
                        query_geom_transformed = shapely.geometry.mapping(
                            query_box_transformed
                        )

                    data, _ = rasterio.mask.mask(
                        f, [query_geom_transformed], crop=True, all_touched=True
                    )

                if layer in [
                    'naip',
                    'buildings',
                    'roads',
                    'waterways',
                    'waterbodies',
                    'water',
                ]:
                    sample['image'].append(data)
                elif layer in ['prior', 'prior_no_osm_no_buildings']:
                    if self.prior_as_input:
                        sample['image'].append(data)
                    else:
                        sample['mask'].append(data)
                elif layer in ['lc']:
                    data = self.raw_enviroatlas_to_idx_map[data]
                    sample['mask'].append(data)
        else:
            raise IndexError(f'query: {query} spans multiple tiles which is not valid')

        sample['image'] = np.concatenate(sample['image'], axis=0)
        sample['mask'] = np.concatenate(sample['mask'], axis=0)

        sample['image'] = torch.from_numpy(sample['image'])
        sample['mask'] = torch.from_numpy(sample['mask'])

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""

        def exists(filename: Path) -> bool:
            return os.path.exists(os.path.join(self.root, 'enviroatlas_lotp', filename))

        # Check if the extracted files already exist
        if all(map(exists, self._files)):
            return

        # Check if the zip files have already been downloaded
        if os.path.exists(os.path.join(self.root, self.filename)):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        download_url(self.url, self.root, filename=self.filename, md5=self.md5)

    def _extract(self) -> None:
        """Extract the dataset."""
        extract_archive(os.path.join(self.root, self.filename))

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Note: only plots the "naip" and "lc" layers.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            ValueError: if the NAIP layer isn't included in ``self.layers``
        """
        if 'naip' not in self.layers or 'lc' not in self.layers:
            raise ValueError("The 'naip' and 'lc' layers must be included for plotting")

        image_layers = []
        mask_layers = []
        for layer in self.layers:
            if layer in [
                'naip',
                'buildings',
                'roads',
                'waterways',
                'waterbodies',
                'water',
            ]:
                image_layers.append(layer)
            elif layer in ['prior', 'prior_no_osm_no_buildings']:
                if self.prior_as_input:
                    image_layers.append(layer)
                else:
                    mask_layers.append(layer)
            elif layer in ['lc']:
                mask_layers.append(layer)

        naip_index = image_layers.index('naip')
        lc_index = mask_layers.index('lc')

        image = np.rollaxis(
            sample['image'][naip_index : naip_index + 3, :, :].numpy(), 0, 3
        )
        mask = sample['mask'][lc_index].numpy()

        num_panels = 2
        showing_predictions = 'prediction' in sample
        if showing_predictions:
            predictions = sample['prediction'].numpy()
            num_panels += 1

        fig, axs = plt.subplots(1, num_panels, figsize=(num_panels * 4, 5))
        axs[0].imshow(image)
        axs[0].axis('off')
        axs[1].imshow(
            mask, vmin=0, vmax=10, cmap=self.highres_cmap, interpolation='none'
        )
        axs[1].axis('off')
        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('Mask')

        if showing_predictions:
            axs[2].imshow(
                predictions,
                vmin=0,
                vmax=10,
                cmap=self.highres_cmap,
                interpolation='none',
            )
            axs[2].axis('off')
            if show_titles:
                axs[2].set_title('Predictions')

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SpaceNet datamodules."""

from typing import Any

import kornia.augmentation as K
import torch
from torch import Tensor
from torch.utils.data import random_split

from ..datasets import SpaceNet1
from ..transforms import AugmentationSequential
from .geo import NonGeoDataModule


class SpaceNet1DataModule(NonGeoDataModule):
    """LightningDataModule implementation for the SpaceNet1 dataset.

    Randomly splits into train/val/test.

    .. versionadded:: 0.4
    """

    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 0,
        val_split_pct: float = 0.1,
        test_split_pct: float = 0.2,
        **kwargs: Any,
    ) -> None:
        """Initialize a new SpaceNet1DataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            val_split_pct: Percentage of the dataset to use as a validation set.
            test_split_pct: Percentage of the dataset to use as a test set.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.SpaceNet1`.
        """
        super().__init__(SpaceNet1, batch_size, num_workers, **kwargs)

        self.val_split_pct = val_split_pct
        self.test_split_pct = test_split_pct

        self.train_aug = AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.PadTo((448, 448)),
            K.RandomRotation(p=0.5, degrees=90),
            K.RandomHorizontalFlip(p=0.5),
            K.RandomVerticalFlip(p=0.5),
            K.RandomSharpness(p=0.5),
            K.ColorJitter(p=0.5, brightness=0.1, contrast=0.1, saturation=0.1, hue=0.1),
            data_keys=['image', 'mask'],
        )
        self.aug = AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.PadTo((448, 448)),
            data_keys=['image', 'mask'],
        )

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.dataset = SpaceNet1(**self.kwargs)
        generator = torch.Generator().manual_seed(0)
        self.train_dataset, self.val_dataset, self.test_dataset = random_split(
            self.dataset,
            [
                1 - self.val_split_pct - self.test_split_pct,
                self.val_split_pct,
                self.test_split_pct,
            ],
            generator,
        )

    def on_after_batch_transfer(
        self, batch: dict[str, Tensor], dataloader_idx: int
    ) -> dict[str, Tensor]:
        """Apply batch augmentations to the batch after it is transferred to the device.

        Args:
            batch: A batch of data that needs to be altered or augmented.
            dataloader_idx: The index of the dataloader to which the batch belongs.

        Returns:
            A batch of data.
        """
        # We add 1 to the mask to map the current {background, building} labels to
        # the values {1, 2}. This is necessary because we add 0 padding to the
        # mask that we want to ignore in the loss function.
        batch['mask'] += 1

        return super().on_after_batch_transfer(batch, dataloader_idx)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""IDTReeS dataset."""

import glob
import os
from collections.abc import Callable
from typing import Any, ClassVar, cast, overload

import fiona
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from rasterio.enums import Resampling
from torch import Tensor
from torchvision.ops import clip_boxes_to_image, remove_small_boxes
from torchvision.utils import draw_bounding_boxes

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, extract_archive, lazy_import


class IDTReeS(NonGeoDataset):
    """IDTReeS dataset.

    The `IDTReeS <https://idtrees.org/competition/>`__
    dataset is a dataset for tree crown detection.

    Dataset features:

    * RGB Image, Canopy Height Model (CHM), Hyperspectral Image (HSI), LiDAR Point Cloud
    * Remote sensing and field data generated by the
      `National Ecological Observatory Network (NEON) <https://data.neonscience.org/>`_
    * 0.1 - 1m resolution imagery
    * Task 1 - object detection (tree crown delination)
    * Task 2 - object classification (species classification)
    * Train set contains 85 images
    * Test set (task 1) contains 153 images
    * Test set (task 2) contains 353 images and tree crown polygons

    Dataset format:

    * optical - three-channel RGB 200x200 geotiff
    * canopy height model - one-channel 20x20 geotiff
    * hyperspectral - 369-channel 20x20 geotiff
    * point cloud - Nx3 LAS file (.las), some files contain RGB colors per point
    * shapely files (.shp) containing polygons
    * csv file containing species labels and other metadata for each polygon

    Dataset classes:

    0. ACPE
    1. ACRU
    2. ACSA3
    3. AMLA
    4. BETUL
    5. CAGL8
    6. CATO6
    7. FAGR
    8. GOLA
    9. LITU
    10. LYLU3
    11. MAGNO
    12. NYBI
    13. NYSY
    14. OXYDE
    15. PEPA37
    16. PIEL
    17. PIPA2
    18. PINUS
    19. PITA
    20. PRSE2
    21. QUAL
    22. QUCO2
    23. QUGE2
    24. QUHE2
    25. QULA2
    26. QULA3
    27. QUMO4
    28. QUNI
    29. QURU
    30. QUERC
    31. ROPS
    32. TSCA

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1101/2021.08.06.453503

    This dataset requires the following additional libraries to be installed:

       * `laspy <https://pypi.org/project/laspy/>`_ to read lidar point clouds
       * `pyvista <https://pypi.org/project/pyvista/>`_ to plot lidar point clouds

    .. versionadded:: 0.2
    """

    classes: ClassVar[dict[str, str]] = {
        'ACPE': 'Acer pensylvanicum L.',
        'ACRU': 'Acer rubrum L.',
        'ACSA3': 'Acer saccharum Marshall',
        'AMLA': 'Amelanchier laevis Wiegand',
        'BETUL': 'Betula sp.',
        'CAGL8': 'Carya glabra (Mill.) Sweet',
        'CATO6': 'Carya tomentosa (Lam.) Nutt.',
        'FAGR': 'Fagus grandifolia Ehrh.',
        'GOLA': 'Gordonia lasianthus (L.) Ellis',
        'LITU': 'Liriodendron tulipifera L.',
        'LYLU3': 'Lyonia lucida (Lam.) K. Koch',
        'MAGNO': 'Magnolia sp.',
        'NYBI': 'Nyssa biflora Walter',
        'NYSY': 'Nyssa sylvatica Marshall',
        'OXYDE': 'Oxydendrum sp.',
        'PEPA37': 'Persea palustris (Raf.) Sarg.',
        'PIEL': 'Pinus elliottii Engelm.',
        'PIPA2': 'Pinus palustris Mill.',
        'PINUS': 'Pinus sp.',
        'PITA': 'Pinus taeda L.',
        'PRSE2': 'Prunus serotina Ehrh.',
        'QUAL': 'Quercus alba L.',
        'QUCO2': 'Quercus coccinea',
        'QUGE2': 'Quercus geminata Small',
        'QUHE2': 'Quercus hemisphaerica W. Bartram ex Willd.',
        'QULA2': 'Quercus laevis Walter',
        'QULA3': 'Quercus laurifolia Michx.',
        'QUMO4': 'Quercus montana Willd.',
        'QUNI': 'Quercus nigra L.',
        'QURU': 'Quercus rubra L.',
        'QUERC': 'Quercus sp.',
        'ROPS': 'Robinia pseudoacacia L.',
        'TSCA': 'Tsuga canadensis (L.) Carriere',
    }
    metadata: ClassVar[dict[str, dict[str, str]]] = {
        'train': {
            'url': 'https://zenodo.org/records/3934932/files/IDTREES_competition_train_v2.zip?download=1',
            'md5': '5ddfa76240b4bb6b4a7861d1d31c299c',
            'filename': 'IDTREES_competition_train_v2.zip',
        },
        'test': {
            'url': 'https://zenodo.org/records/3934932/files/IDTREES_competition_test_v2.zip?download=1',
            'md5': 'b108931c84a70f2a38a8234290131c9b',
            'filename': 'IDTREES_competition_test_v2.zip',
        },
    }
    directories: ClassVar[dict[str, list[str]]] = {
        'train': ['train'],
        'test': ['task1', 'task2'],
    }
    image_size = (200, 200)

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        task: str = 'task1',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new IDTReeS dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train" or "test"
            task: 'task1' for detection, 'task2' for detection + classification
                (only relevant for split='test')
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If laspy is not installed.
        """
        lazy_import('laspy')

        assert split in ['train', 'test']
        assert task in ['task1', 'task2']

        self.root = root
        self.split = split
        self.task = task
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.class2idx = {c: i for i, c in enumerate(self.classes)}
        self.idx2class = {i: c for i, c in enumerate(self.classes)}
        self.num_classes = len(self.classes)
        self._verify()
        self.images, self.geometries, self.labels = self._load(root)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        path = self.images[index]
        image = self._load_image(path).to(torch.uint8)
        hsi = self._load_image(path.replace('RGB', 'HSI'))
        chm = self._load_image(path.replace('RGB', 'CHM'))
        las = self._load_las(path.replace('RGB', 'LAS').replace('.tif', '.las'))
        sample = {'image': image, 'hsi': hsi, 'chm': chm, 'las': las}

        if self.split == 'test':
            if self.task == 'task2':
                sample['boxes'] = self._load_boxes(path)
                h, w = sample['image'].shape[1:]
                sample['boxes'], _ = self._filter_boxes(
                    image_size=(h, w), min_size=1, boxes=sample['boxes'], labels=None
                )
        else:
            sample['boxes'] = self._load_boxes(path)
            sample['label'] = self._load_target(path)

            h, w = sample['image'].shape[1:]
            sample['boxes'], sample['label'] = self._filter_boxes(
                image_size=(h, w),
                min_size=1,
                boxes=sample['boxes'],
                labels=sample['label'],
            )

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.images)

    def _load_image(self, path: Path) -> Tensor:
        """Load a tiff file.

        Args:
            path: path to .tif file

        Returns:
            the image
        """
        with rasterio.open(path) as f:
            array = f.read(out_shape=self.image_size, resampling=Resampling.bilinear)
        tensor = torch.from_numpy(array)
        return tensor

    def _load_las(self, path: Path) -> Tensor:
        """Load a single point cloud.

        Args:
            path: path to .las file

        Returns:
            the point cloud
        """
        laspy = lazy_import('laspy')
        las = laspy.read(path)
        array: np.typing.NDArray[np.int_] = np.stack([las.x, las.y, las.z], axis=0)
        tensor = torch.from_numpy(array)
        return tensor

    def _load_boxes(self, path: Path) -> Tensor:
        """Load object bounding boxes.

        Args:
            path: path to .tif file

        Returns:
            the bounding boxes
        """
        base_path = os.path.basename(path)
        geometries = cast(dict[int, dict[str, Any]], self.geometries)

        # Find object ids and geometries
        # The train set geometry->image mapping is contained
        # in the train/Field/itc_rsFile.csv file
        if self.split == 'train':
            indices = self.labels['rsFile'] == base_path
            ids = self.labels[indices]['id'].tolist()
            geoms = [geometries[i]['geometry']['coordinates'][0][:4] for i in ids]
        # The test set has no mapping csv. The mapping is inside of the geometry
        # properties i.e. geom["property"]["plotID"] contains the RGB image filename
        # Return all geometries with the matching RGB image filename of the sample
        else:
            ids = [
                k
                for k, v in geometries.items()
                if v['properties']['plotID'] == base_path
            ]
            geoms = [geometries[i]['geometry']['coordinates'][0][:4] for i in ids]

        # Convert to pixel coords
        boxes = []
        with rasterio.open(path) as f:
            for geom in geoms:
                coords = [f.index(x, y) for x, y in geom]
                xmin = min(coord[1] for coord in coords)
                xmax = max(coord[1] for coord in coords)
                ymin = min(coord[0] for coord in coords)
                ymax = max(coord[0] for coord in coords)
                boxes.append([xmin, ymin, xmax, ymax])

        tensor = torch.tensor(boxes)
        return tensor

    def _load_target(self, path: Path) -> Tensor:
        """Load target label for a single sample.

        Args:
            path: path to image

        Returns:
            the label
        """
        # Find indices for objects in the image
        base_path = os.path.basename(path)
        indices = self.labels['rsFile'] == base_path

        # Load object labels
        classes = self.labels[indices]['taxonID'].tolist()
        labels = [self.class2idx[c] for c in classes]
        tensor = torch.tensor(labels)
        return tensor

    def _load(
        self, root: Path
    ) -> tuple[list[str], dict[int, dict[str, Any]] | None, Any]:
        """Load files, geometries, and labels.

        Args:
            root: root directory

        Returns:
            the image path, geometries, and labels
        """
        if self.split == 'train':
            directory = os.path.join(root, self.directories[self.split][0])
            labels: pd.DataFrame = self._load_labels(directory)
            geoms = self._load_geometries(directory)
        else:
            directory = os.path.join(root, self.task)
            if self.task == 'task1':
                geoms = None
                labels = None
            else:
                geoms = self._load_geometries(directory)
                labels = None

        images = glob.glob(os.path.join(directory, 'RemoteSensing', 'RGB', '*.tif'))

        return images, geoms, labels

    def _load_labels(self, directory: Path) -> Any:
        """Load the csv files containing the labels.

        Args:
            directory: directory containing csv files

        Returns:
            a pandas DataFrame containing the labels for each image
        """
        path_mapping = os.path.join(directory, 'Field', 'itc_rsFile.csv')
        path_labels = os.path.join(directory, 'Field', 'train_data.csv')
        df_mapping = pd.read_csv(path_mapping)
        df_labels = pd.read_csv(path_labels)
        df_mapping = df_mapping.set_index('indvdID', drop=True)
        df_labels = df_labels.set_index('indvdID', drop=True)
        df = df_labels.join(df_mapping, on='indvdID')
        df = df.drop_duplicates()
        df.reset_index()
        return df

    def _load_geometries(self, directory: Path) -> dict[int, dict[str, Any]]:
        """Load the shape files containing the geometries.

        Args:
            directory: directory containing .shp files

        Returns:
            a dict containing the geometries for each object
        """
        filepaths = glob.glob(os.path.join(directory, 'ITC', '*.shp'))

        i = 0
        features: dict[int, dict[str, Any]] = {}
        for path in filepaths:
            with fiona.open(path) as src:
                for feature in src:
                    # The train set has a unique id for each geometry in the properties
                    if self.split == 'train':
                        features[feature['properties']['id']] = feature
                    # The test set has no unique id so create a dummy id
                    else:
                        features[i] = feature
                        i += 1
        return features

    @overload
    def _filter_boxes(
        self, image_size: tuple[int, int], min_size: int, boxes: Tensor, labels: Tensor
    ) -> tuple[Tensor, Tensor]: ...

    @overload
    def _filter_boxes(
        self, image_size: tuple[int, int], min_size: int, boxes: Tensor, labels: None
    ) -> tuple[Tensor, None]: ...

    def _filter_boxes(
        self,
        image_size: tuple[int, int],
        min_size: int,
        boxes: Tensor,
        labels: Tensor | None,
    ) -> tuple[Tensor, Tensor | None]:
        """Clip boxes to image size and filter boxes with sides less than ``min_size``.

        Args:
            image_size: tuple of (height, width) of image
            min_size: filter boxes that have any side less than min_size
            boxes: [N, 4] shape tensor of xyxy bounding box coordinates
            labels: (Optional) [N,] shape tensor of bounding box labels

        Returns:
            a tuple of filtered boxes and labels
        """
        boxes = clip_boxes_to_image(boxes=boxes, size=image_size)
        indices = remove_small_boxes(boxes=boxes, min_size=min_size)

        boxes = boxes[indices]
        if labels is not None:
            labels = labels[indices]

        return boxes, labels

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        url = self.metadata[self.split]['url']
        md5 = self.metadata[self.split]['md5']
        filename = self.metadata[self.split]['filename']
        directories = self.directories[self.split]

        # Check if the files already exist
        exists = [
            os.path.exists(os.path.join(self.root, directory))
            for directory in directories
        ]
        if all(exists):
            return

        # Check if zip file already exists (if so then extract)
        filepath = os.path.join(self.root, filename)
        if os.path.exists(filepath):
            extract_archive(filepath)
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        download_url(
            url, self.root, filename=filename, md5=md5 if self.checksum else None
        )
        filepath = os.path.join(self.root, filename)
        extract_archive(filepath)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
        hsi_indices: tuple[int, int, int] = (0, 1, 2),
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle
            hsi_indices: tuple of indices to create HSI false color image

        Returns:
            a matplotlib Figure with the rendered sample
        """
        assert len(hsi_indices) == 3

        def normalize(x: Tensor) -> Tensor:
            return (x - x.min()) / (x.max() - x.min())

        ncols = 3

        hsi = normalize(sample['hsi'][hsi_indices, :, :]).permute((1, 2, 0)).numpy()
        chm = normalize(sample['chm']).permute((1, 2, 0)).numpy()

        if 'boxes' in sample and len(sample['boxes']):
            labels = (
                [self.idx2class[int(i)] for i in sample['label']]
                if 'label' in sample
                else None
            )
            image = draw_bounding_boxes(
                image=sample['image'], boxes=sample['boxes'], labels=labels
            )
            image = image.permute((1, 2, 0)).numpy()
        else:
            image = sample['image'].permute((1, 2, 0)).numpy()

        if 'prediction_boxes' in sample and len(sample['prediction_boxes']):
            ncols += 1
            labels = (
                [self.idx2class[int(i)] for i in sample['prediction_label']]
                if 'prediction_label' in sample
                else None
            )
            preds = draw_bounding_boxes(
                image=sample['image'], boxes=sample['prediction_boxes'], labels=labels
            )
            preds = preds.permute((1, 2, 0)).numpy()

        fig, axs = plt.subplots(ncols=ncols, figsize=(ncols * 10, 10))
        axs[0].imshow(image)
        axs[0].axis('off')
        axs[1].imshow(hsi)
        axs[1].axis('off')
        axs[2].imshow(chm)
        axs[2].axis('off')
        if ncols > 3:
            axs[3].imshow(preds)
            axs[3].axis('off')

        if show_titles:
            axs[0].set_title('Ground Truth')
            axs[1].set_title('Hyperspectral False Color Image')
            axs[2].set_title('Canopy Height Model')
            if ncols > 3:
                axs[3].set_title('Predictions')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig

    def plot_las(self, index: int) -> 'pyvista.Plotter':  # type: ignore[name-defined] # noqa: F821
        """Plot a sample point cloud at the index.

        Args:
            index: index to plot

        Returns:
            pyvista.PolyData object. Run pyvista.plot(point_cloud, ...) to display

        Raises:
            DependencyNotFoundError: If laspy or pyvista are not installed.

        .. versionchanged:: 0.4
           Ported from Open3D to PyVista, *colormap* parameter removed.
        """
        laspy = lazy_import('laspy')
        pyvista = lazy_import('pyvista')
        path = self.images[index]
        path = path.replace('RGB', 'LAS').replace('.tif', '.las')
        las = laspy.read(path)
        points: np.typing.NDArray[np.int_] = np.stack(
            [las.x, las.y, las.z], axis=0
        ).transpose((1, 0))
        point_cloud = pyvista.PolyData(points)

        # Some point cloud files have no color->points mapping
        if hasattr(las, 'red'):
            colors = np.stack([las.red, las.green, las.blue], axis=0)
            colors = colors.transpose((1, 0)) / np.iinfo(np.uint16).max
            point_cloud['colors'] = colors

        return point_cloud



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Common trainer utilities."""

import warnings
from collections import OrderedDict
from typing import cast

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.modules import Conv2d, Module


def extract_backbone(path: str) -> tuple[str, 'OrderedDict[str, Tensor]']:
    """Extracts a backbone from a lightning checkpoint file.

    Args:
        path: path to checkpoint file (.ckpt)

    Returns:
        tuple containing model name and state dict

    Raises:
        ValueError: if 'model' or 'backbone' not in
            checkpoint['hyper_parameters']

    .. versionchanged:: 0.4
        Renamed from *extract_encoder* to *extract_backbone*
    """
    checkpoint = torch.load(path, map_location=torch.device('cpu'))
    if 'model' in checkpoint['hyper_parameters']:
        name = checkpoint['hyper_parameters']['model']
        state_dict = checkpoint['state_dict']
        state_dict = OrderedDict({k: v for k, v in state_dict.items() if 'model.' in k})
        state_dict = OrderedDict(
            {k.replace('model.', ''): v for k, v in state_dict.items()}
        )
    elif 'backbone' in checkpoint['hyper_parameters']:
        name = checkpoint['hyper_parameters']['backbone']
        state_dict = checkpoint['state_dict']
        state_dict = OrderedDict(
            {k: v for k, v in state_dict.items() if 'model.backbone.model' in k}
        )
        state_dict = OrderedDict(
            {k.replace('model.backbone.model.', ''): v for k, v in state_dict.items()}
        )
    else:
        raise ValueError(
            'Unknown checkpoint task. Only backbone or model extraction is supported'
        )

    return name, state_dict


def _get_input_layer_name_and_module(model: Module) -> tuple[str, Module]:
    """Retrieve the input layer name and module from a timm model.

    Args:
        model: timm model
    """
    keys = []
    children = list(model.named_children())
    while children != []:
        name, module = children[0]
        keys.append(name)
        children = list(module.named_children())

    key = '.'.join(keys)
    return key, module


def load_state_dict(
    model: Module, state_dict: 'OrderedDict[str, Tensor]'
) -> tuple[list[str], list[str]]:
    """Load pretrained resnet weights to a model.

    Args:
        model: model to load the pretrained weights to
        state_dict: dict containing tensor parameters

    Returns:
        The missing and unexpected keys

    Warns:
        If input channels in model != pretrained model input channels
        If num output classes in model != pretrained model num classes
    """
    input_module_key, input_module = _get_input_layer_name_and_module(model)
    in_channels = input_module.in_channels
    expected_in_channels = state_dict[input_module_key + '.weight'].shape[1]

    output_module_key, output_module = list(model.named_children())[-1]
    if isinstance(output_module, nn.Identity):
        num_classes = model.num_features
    else:
        num_classes = output_module.out_features
    expected_num_classes = None
    if output_module_key + '.weight' in state_dict:
        expected_num_classes = state_dict[output_module_key + '.weight'].shape[0]

    if in_channels != expected_in_channels:
        warnings.warn(
            f'input channels {in_channels} != input channels in pretrained'
            f' model {expected_in_channels}. Overriding with new input channels'
        )
        del state_dict[input_module_key + '.weight']

    if expected_num_classes and num_classes != expected_num_classes:
        warnings.warn(
            f'num classes {num_classes} != num classes in pretrained model'
            f' {expected_num_classes}. Overriding with new num classes'
        )
        del (
            state_dict[output_module_key + '.weight'],
            state_dict[output_module_key + '.bias'],
        )

    missing_keys: list[str]
    unexpected_keys: list[str]
    missing_keys, unexpected_keys = model.load_state_dict(state_dict, strict=False)
    return missing_keys, unexpected_keys


def reinit_initial_conv_layer(
    layer: Conv2d,
    new_in_channels: int,
    keep_rgb_weights: bool,
    new_stride: int | tuple[int, int] | None = None,
    new_padding: str | int | tuple[int, int] | None = None,
) -> Conv2d:
    """Clones a Conv2d layer while optionally retaining some of the original weights.

    When replacing the first convolutional layer in a model with one that operates over
    different number of input channels, we sometimes want to keep a subset of the kernel
    weights the same (e.g. the RGB weights of an ImageNet pretrained model). This is a
    convenience function that performs that function.

    Args:
        layer: the Conv2d layer to initialize
        new_in_channels: the new number of input channels
        keep_rgb_weights: flag indicating whether to re-initialize the first 3 channels
        new_stride: optionally, overwrites the ``layer``'s stride with this value
        new_padding: optionally, overwrites the ``layers``'s padding with this value

    Returns:
        a Conv2d layer with new kernel weights
    """
    use_bias = layer.bias is not None
    if keep_rgb_weights:
        w_old = layer.weight.data[:, :3, :, :].clone()
        if use_bias:
            b_old = cast(Tensor, layer.bias).data.clone()

    updated_stride = layer.stride if new_stride is None else new_stride
    updated_padding = layer.padding if new_padding is None else new_padding

    new_layer = Conv2d(
        new_in_channels,
        layer.out_channels,
        kernel_size=layer.kernel_size,  # type: ignore[arg-type]
        stride=updated_stride,  # type: ignore[arg-type]
        padding=updated_padding,  # type: ignore[arg-type]
        dilation=layer.dilation,  # type: ignore[arg-type]
        groups=layer.groups,
        bias=use_bias,
        padding_mode=layer.padding_mode,
    )
    nn.init.kaiming_normal_(new_layer.weight, mode='fan_out', nonlinearity='relu')

    if keep_rgb_weights:
        new_layer.weight.data[:, :3, :, :] = w_old
        if use_bias:
            cast(Tensor, new_layer.bias).data = b_old

    return new_layer



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Landsat datasets."""

import abc
from collections.abc import Callable, Iterable, Sequence
from typing import Any

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import RGBBandsMissingError
from .geo import RasterDataset
from .utils import Path


class Landsat(RasterDataset, abc.ABC):
    """Abstract base class for all Landsat datasets.

    `Landsat <https://landsat.gsfc.nasa.gov/>`__ is a joint NASA/USGS program,
    providing the longest continuous space-based record of Earth's land in existence.

    If you use this dataset in your research, please cite it using the following format:

    * https://www.usgs.gov/centers/eros/data-citation

    If you use any of the following Level-2 products, there may be additional citation
    requirements, including papers you can cite. See the "Citation Information" section
    of the following pages:

    * `Surface Temperature <https://www.usgs.gov/landsat-missions/landsat-collection-2-surface-temperature>`_
    * `Surface Reflectance <https://www.usgs.gov/landsat-missions/landsat-collection-2-surface-reflectance>`_
    * `U.S. Analysis Ready Data <https://www.usgs.gov/landsat-missions/landsat-collection-2-us-analysis-ready-data>`_
    """

    # https://www.usgs.gov/landsat-missions/landsat-collection-2
    filename_regex = r"""
        ^L
        (?P<sensor>[COTEM])
        (?P<satellite>\d{2})
        _(?P<processing_correction_level>[A-Z0-9]{4})
        _(?P<wrs_path>\d{3})
        (?P<wrs_row>\d{3})
        _(?P<date>\d{8})
        _(?P<processing_date>\d{8})
        _(?P<collection_number>\d{2})
        _(?P<collection_category>[A-Z0-9]{2})
        _(?P<band>[A-Z0-9_]+)
        \.
    """

    separate_files = True

    @property
    @abc.abstractmethod
    def default_bands(self) -> tuple[str, ...]:
        """Bands to load by default."""

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | None = None,
        bands: Sequence[str] | None = None,
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS
                (defaults to the resolution of the first file found)
            bands: bands to return (defaults to all bands)
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.

        .. versionchanged:: 0.5
           *root* was renamed to *paths*.
        """
        bands = bands or self.default_bands
        self.filename_glob = self.filename_glob.format(bands[0])

        super().__init__(paths, crs, res, bands, transforms, cache)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.

        .. versionchanged:: 0.3
           Method now takes a sample dict, not a Tensor. Additionally, possible to
           show subplot titles and/or use a custom suptitle.
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        image = sample['image'][rgb_indices].permute(1, 2, 0).float()

        # Stretch to the full range
        image = (image - image.min()) / (image.max() - image.min())

        fig, ax = plt.subplots(1, 1, figsize=(4, 4))

        ax.imshow(image)
        ax.axis('off')

        if show_titles:
            ax.set_title('Image')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig


class Landsat1(Landsat):
    """Landsat 1 Multispectral Scanner (MSS)."""

    filename_glob = 'LM01_*_{}.*'

    default_bands = ('B4', 'B5', 'B6', 'B7')
    rgb_bands = ('B6', 'B5', 'B4')


class Landsat2(Landsat1):
    """Landsat 2 Multispectral Scanner (MSS)."""

    filename_glob = 'LM02_*_{}.*'


class Landsat3(Landsat1):
    """Landsat 3 Multispectral Scanner (MSS)."""

    filename_glob = 'LM03_*_{}.*'


class Landsat4MSS(Landsat):
    """Landsat 4 Multispectral Scanner (MSS)."""

    filename_glob = 'LM04_*_{}.*'

    default_bands = ('B1', 'B2', 'B3', 'B4')
    rgb_bands = ('B3', 'B2', 'B1')


class Landsat4TM(Landsat):
    """Landsat 4 Thematic Mapper (TM)."""

    filename_glob = 'LT04_*_{}.*'

    default_bands = ('SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7')
    rgb_bands = ('SR_B3', 'SR_B2', 'SR_B1')


class Landsat5MSS(Landsat4MSS):
    """Landsat 4 Multispectral Scanner (MSS)."""

    filename_glob = 'LM04_*_{}.*'


class Landsat5TM(Landsat4TM):
    """Landsat 5 Thematic Mapper (TM)."""

    filename_glob = 'LT05_*_{}.*'


class Landsat7(Landsat):
    """Landsat 7 Enhanced Thematic Mapper Plus (ETM+)."""

    filename_glob = 'LE07_*_{}.*'

    default_bands = ('SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7')
    rgb_bands = ('SR_B3', 'SR_B2', 'SR_B1')


class Landsat8(Landsat):
    """Landsat 8 Operational Land Imager (OLI) and Thermal Infrared Sensor (TIRS)."""

    filename_glob = 'LC08_*_{}.*'

    default_bands = ('SR_B1', 'SR_B2', 'SR_B3', 'SR_B4', 'SR_B5', 'SR_B6', 'SR_B7')
    rgb_bands = ('SR_B4', 'SR_B3', 'SR_B2')


class Landsat9(Landsat8):
    """Landsat 9 Operational Land Imager (OLI-2) and Thermal Infrared Sensor (TIRS-2)."""

    filename_glob = 'LC09_*_{}.*'



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""SustainBench Crop Yield dataset."""

import os
from collections.abc import Callable
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, extract_archive


class SustainBenchCropYield(NonGeoDataset):
    """SustainBench Crop Yield Dataset.

    This dataset contains MODIS band histograms and soybean yield
    estimates for selected counties in the USA, Argentina and Brazil.
    The dataset is part of the
    `SustainBench <https://sustainlab-group.github.io/sustainbench/docs/datasets/sdg2/crop_yield.html>`_
    datasets for tackling the UN Sustainable Development Goals (SDGs).

    Dataset Format:

    * .npz files of stacked samples

    Dataset Features:

    * input histogram of 7 surface reflectance and 2 surface temperature
      bands from MODIS pixel values in 32 ranges across 32 timesteps
      resulting in 32x32x9 input images
    * regression target value of soybean yield in metric tonnes per
      harvested hectare

    If you use this dataset in your research, please cite:

    * https://doi.org/10.1145/3209811.3212707
    * https://doi.org/10.1609/aaai.v31i1.11172

    .. versionadded:: 0.5
    """

    valid_countries = ('usa', 'brazil', 'argentina')

    md5 = '362bad07b51a1264172b8376b39d1fc9'

    url = 'https://drive.google.com/file/d/1lhbmICpmNuOBlaErywgiD6i9nHuhuv0A/view?usp=drive_link'

    dir = 'soybeans'

    valid_splits = ('train', 'dev', 'test')

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        countries: list[str] = ['usa'],
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "dev", or "test"
            countries: which countries to include in the dataset
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 after downloading files (may be slow)

        Raises:
            AssertionError: if ``countries`` contains invalid countries or if ``split``
                is invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert set(countries).issubset(
            self.valid_countries
        ), f'Please choose a subset of these valid countried: {self.valid_countries}.'
        self.countries = countries

        assert (
            split in self.valid_splits
        ), f'Pleas choose one of these valid data splits {self.valid_splits}.'
        self.split = split

        self.root = root
        self.transforms = transforms
        self.download = download
        self.checksum = checksum

        self._verify()

        self.images = []
        self.features = []

        for country in self.countries:
            image_file_path = os.path.join(
                self.root, self.dir, country, f'{self.split}_hists.npz'
            )
            target_file_path = image_file_path.replace('_hists', '_yields')
            years_file_path = image_file_path.replace('_hists', '_years')
            ndvi_file_path = image_file_path.replace('_hists', '_ndvi')

            npz_file = np.load(image_file_path)['data']
            target_npz_file = np.load(target_file_path)['data']
            year_npz_file = np.load(years_file_path)['data']
            ndvi_npz_file = np.load(ndvi_file_path)['data']
            num_data_points = npz_file.shape[0]
            for idx in range(num_data_points):
                sample = npz_file[idx]
                sample = torch.from_numpy(sample).permute(2, 0, 1).to(torch.float32)
                self.images.append(sample)

                target = target_npz_file[idx]
                year = year_npz_file[idx]
                ndvi = ndvi_npz_file[idx]

                features = {
                    'label': torch.tensor(target).to(torch.float32),
                    'year': torch.tensor(int(year)),
                    'ndvi': torch.from_numpy(ndvi).to(dtype=torch.float32),
                }
                self.features.append(features)

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.images)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            data and label at that index
        """
        sample: dict[str, Tensor] = {'image': self.images[index]}
        sample.update(self.features[index])

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        pathname = os.path.join(self.root, self.dir)
        if os.path.exists(pathname):
            return

        # Check if the zip files have already been downloaded
        pathname = os.path.join(self.root, self.dir) + '.zip'
        if os.path.exists(pathname):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset and extract it."""
        download_url(
            self.url,
            self.root,
            filename=self.dir + '.zip',
            md5=self.md5 if self.checksum else None,
        )
        self._extract()

    def _extract(self) -> None:
        """Extract the dataset."""
        zipfile_path = os.path.join(self.root, self.dir) + '.zip'
        extract_archive(zipfile_path, self.root)

    def plot(
        self,
        sample: dict[str, Any],
        band_idx: int = 0,
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample return by :meth:`__getitem__`
            band_idx: which of the nine histograms to index
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample

        """
        image, label = sample['image'], sample['label'].item()

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = sample['prediction'].item()

        fig, ax = plt.subplots(1, 1, figsize=(10, 10))

        ax.imshow(image.permute(1, 2, 0)[:, :, band_idx])
        ax.axis('off')

        if show_titles:
            title = f'Label: {label:.3f}'
            if showing_predictions:
                title += f'\nPrediction: {prediction:.3f}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Dataset-specific exceptions."""

from torch.utils.data import Dataset


class DatasetNotFoundError(FileNotFoundError):
    """Raised when a dataset is requested but doesn't exist.

    .. versionadded:: 0.6
    """

    def __init__(self, dataset: Dataset[object]) -> None:
        """Initialize a new DatasetNotFoundError instance.

        Args:
            dataset: The dataset that was requested.
        """
        msg = 'Dataset not found'

        if hasattr(dataset, 'root'):
            var = 'root'
            val = dataset.root
        elif hasattr(dataset, 'paths'):
            var = 'paths'
            val = dataset.paths
        else:
            super().__init__(f'{msg}.')
            return

        msg += f' in `{var}={val!r}` and '

        if hasattr(dataset, 'download') and not dataset.download:
            msg += '`download=False`'
        else:
            msg += 'cannot be automatically downloaded'

        msg += f', either specify a different `{var}` or '

        if hasattr(dataset, 'download') and not dataset.download:
            msg += 'use `download=True` to automatically'
        else:
            msg += 'manually'

        msg += ' download the dataset.'

        super().__init__(msg)


class DependencyNotFoundError(Exception):
    """Raised when an optional dataset dependency is not installed.

    .. versionadded:: 0.6
    """


class RGBBandsMissingError(ValueError):
    """Raised when a dataset is missing RGB bands for plotting.

    .. versionadded:: 0.6
    """

    def __init__(self) -> None:
        """Initialize a new RGBBandsMissingError instance."""
        msg = 'Dataset does not contain some of the RGB bands'
        super().__init__(msg)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""UC Merced datamodule."""

from typing import Any

import kornia.augmentation as K

from ..datasets import UCMerced
from ..transforms import AugmentationSequential
from .geo import NonGeoDataModule


class UCMercedDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the UC Merced dataset.

    Uses random train/val/test splits.
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new UCMercedDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.UCMerced`.
        """
        super().__init__(UCMerced, batch_size, num_workers, **kwargs)

        self.aug = AugmentationSequential(
            K.Normalize(mean=self.mean, std=self.std),
            K.Resize(size=256),
            data_keys=['image'],
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""COWC datamodule."""

from typing import Any

from torch import Generator
from torch.utils.data import random_split

from ..datasets import COWCCounting
from .geo import NonGeoDataModule


class COWCCountingDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the COWC Counting dataset."""

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new COWCCountingDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.COWCCounting`.
        """
        super().__init__(COWCCounting, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.dataset = COWCCounting(split='train', **self.kwargs)
        self.test_dataset = COWCCounting(split='test', **self.kwargs)
        self.train_dataset, self.val_dataset = random_split(
            self.dataset,
            [len(self.dataset) - len(self.test_dataset), len(self.test_dataset)],
            generator=Generator().manual_seed(0),
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""South America Soybean Dataset."""

import os
from collections.abc import Callable, Iterable
from typing import Any, ClassVar

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import RasterDataset
from .utils import Path, download_url


class SouthAmericaSoybean(RasterDataset):
    """South America Soybean Dataset.

    This dataset produced annual 30-m soybean maps of South America from 2001 to 2021.

    Link: https://www.nature.com/articles/s41893-021-00729-z

    Dataset contains 2 classes:

    0. other
    1. soybean

    Dataset Format:

    * 21 .tif files


    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1038/s41893-021-00729-z

    .. versionadded:: 0.6
    """

    filename_glob = 'SouthAmerica_Soybean_*.*'
    filename_regex = r'SouthAmerica_Soybean_(?P<year>\d{4})'

    date_format = '%Y'
    is_image = False
    url = 'https://glad.umd.edu/projects/AnnualClassMapsV1/SouthAmerica_Soybean_{}.tif'

    md5s: ClassVar[dict[int, str]] = {
        2021: 'edff3ada13a1a9910d1fe844d28ae4f',
        2020: '0709dec807f576c9707c8c7e183db31',
        2019: '441836493bbcd5e123cff579a58f5a4f',
        2018: '503c2d0a803c2a2629ebbbd9558a3013',
        2017: '4d0487ac1105d171e5f506f1766ea777',
        2016: '770c558f6ac40550d0e264da5e44b3e',
        2015: '6beb96a61fe0e9ce8c06263e500dde8f',
        2014: '824ff91c62a4ba9f4ccfd281729830e5',
        2013: '0263e19b3cae6fdaba4e3b450cef985e',
        2012: '9f3a71097c9836fcff18a13b9ba608b2',
        2011: 'b73352ebea3d5658959e9044ec526143',
        2010: '9264532d36ffa93493735a6e44caef0d',
        2009: '341387c1bb42a15140c80702e4cca02d',
        2008: '96fc3f737ab3ce9bcd16cbf7761427e2',
        2007: 'bb8549b6674163fe20ffd47ec4ce8903',
        2006: 'eabaa525414ecbff89301d3d5c706f0b',
        2005: '89faae27f9b5afbd06935a465e5fe414',
        2004: 'f9882ca9c70e054e50172835cb75a8c3',
        2003: 'cad5ed461ff4ab45c90177841aaecad2',
        2002: '8a4a9dcea54b3ec7de07657b9f2c0893',
        2001: '2914b0af7590a0ca4dfa9ccefc99020f',
    }

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | None = None,
        years: list[int] = [2021],
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS
                (defaults to the resolution of the first file found)
            years: list of years for which to use the South America Soybean layer
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 after downloading files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        self.paths = paths
        self.download = download
        self.checksum = checksum
        self.years = years
        self._verify()

        super().__init__(paths, crs, res, transforms=transforms, cache=cache)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        if self.files:
            return
        assert isinstance(self.paths, str | os.PathLike)

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        for year in self.years:
            download_url(
                self.url.format(year),
                self.paths,
                md5=self.md5s[year] if self.checksum else None,
            )

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        mask = sample['mask'].squeeze()
        ncols = 1

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'].squeeze()
            ncols = 2

        fig, axs = plt.subplots(
            nrows=1, ncols=ncols, figsize=(ncols * 4, 4), squeeze=False
        )

        axs[0, 0].imshow(mask, interpolation='none')
        axs[0, 0].axis('off')

        if show_titles:
            axs[0, 0].set_title('Mask')

        if showing_predictions:
            axs[0, 1].imshow(pred, interpolation='none')
            axs[0, 1].axis('off')
            if show_titles:
                axs[0, 1].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Cloud Cover Detection Challenge dataset."""

import os
from collections.abc import Callable, Sequence
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError, RGBBandsMissingError
from .geo import NonGeoDataset
from .utils import Path, which


class CloudCoverDetection(NonGeoDataset):
    """Sentinel-2 Cloud Cover Segmentation Dataset.

    This training dataset was generated as part of a `crowdsourcing competition
    <https://www.drivendata.org/competitions/83/cloud-cover/>`_ on DrivenData.org, and
    later on was validated using a team of expert annotators. See `this website
    <https://beta.source.coop/radiantearth/cloud-cover-detection-challenge/>`__
    for dataset details.

    The dataset consists of Sentinel-2 satellite imagery and corresponding cloudy
    labels stored as GeoTiffs. There are 22,728 chips in the training data,
    collected between 2018 and 2020.

    Each chip has:

    * 4 multi-spectral bands from Sentinel-2 L2A product. The four bands are
      [B02, B03, B04, B08] (refer to Sentinel-2 documentation for more
      information about the bands).
    * Label raster for the corresponding source tile representing a binary
      classification for if the pixel is a cloud or not.

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.34911/RDNT.HFQ6M7

    .. note::

       This dataset requires the following additional library to be installed:

       * `azcopy <https://github.com/Azure/azure-storage-azcopy>`_: to download the
         dataset from Source Cooperative.

    .. versionadded:: 0.4
    """

    url = 'https://radiantearth.blob.core.windows.net/mlhub/ref_cloud_cover_detection_challenge_v1/final'
    all_bands = ('B02', 'B03', 'B04', 'B08')
    rgb_bands = ('B04', 'B03', 'B02')
    splits: ClassVar[dict[str, str]] = {'train': 'public', 'test': 'private'}

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        bands: Sequence[str] = all_bands,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
    ) -> None:
        """Initiatlize a CloudCoverDetection instance.

        Args:
            root: root directory where dataset can be found
            split: 'train' or 'test'
            bands: the subset of bands to load
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory

        Raises:
            AssertionError: If *split* or *bands* are invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        assert set(bands) <= set(self.all_bands)

        self.root = root
        self.split = split
        self.bands = bands
        self.transforms = transforms
        self.download = download

        self.csv = os.path.join(self.root, self.split, f'{self.split}_metadata.csv')
        self._verify()

        self.metadata = pd.read_csv(self.csv)

    def __len__(self) -> int:
        """Return the number of items in the dataset.

        Returns:
            length of dataset in integer
        """
        return len(self.metadata)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Returns a sample from dataset.

        Args:
            index: index to return

        Returns:
            data and label at given index
        """
        chip_id = self.metadata.iat[index, 0]
        image = self._load_image(chip_id)
        label = self._load_target(chip_id)
        sample = {'image': image, 'mask': label}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_image(self, chip_id: str) -> Tensor:
        """Load all source images for a chip.

        Args:
            chip_id: ID of the chip.

        Returns:
            a tensor of stacked source image data
        """
        path = os.path.join(self.root, self.split, f'{self.split}_features', chip_id)
        images = []
        for band in self.bands:
            with rasterio.open(os.path.join(path, f'{band}.tif')) as src:
                images.append(src.read(1).astype(np.float32))
        return torch.from_numpy(np.stack(images, axis=0))

    def _load_target(self, chip_id: str) -> Tensor:
        """Load label image for a chip.

        Args:
            chip_id: ID of the chip.

        Returns:
            a tensor of the label image data
        """
        path = os.path.join(self.root, self.split, f'{self.split}_labels')
        with rasterio.open(os.path.join(path, f'{chip_id}.tif')) as src:
            return torch.from_numpy(src.read(1).astype(np.int64))

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        if os.path.exists(self.csv):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        directory = os.path.join(self.root, self.split)
        os.makedirs(directory, exist_ok=True)
        url = f'{self.url}/{self.splits[self.split]}'
        azcopy = which('azcopy')
        azcopy('sync', url, directory, '--recursive=true')

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            time_step: time step at which to access image, beginning with 0
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample

        Raises:
            RGBBandsMissingError: If *bands* does not include all RGB bands.
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise RGBBandsMissingError()

        if 'prediction' in sample:
            prediction = sample['prediction']
            n_cols = 3
        else:
            n_cols = 2

        image, mask = sample['image'] / 3000, sample['mask']

        fig, axs = plt.subplots(nrows=1, ncols=n_cols, figsize=(10, n_cols * 5))

        axs[0].imshow(image.permute(1, 2, 0))
        axs[0].axis('off')
        axs[1].imshow(mask)
        axs[1].axis('off')

        if 'prediction' in sample:
            axs[2].imshow(prediction)
            axs[2].axis('off')
            if show_titles:
                axs[2].set_title('Prediction')

        if show_titles:
            axs[0].set_title('Image')
            axs[1].set_title('Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""NASA Marine Debris datamodule."""

from typing import Any

import kornia.augmentation as K
import torch
from torch.utils.data import random_split

from ..datasets import NASAMarineDebris
from ..transforms import AugmentationSequential
from .geo import NonGeoDataModule
from .utils import AugPipe, collate_fn_detection


class NASAMarineDebrisDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the NASA Marine Debris dataset.

    .. versionadded:: 0.2
    """

    std = torch.tensor(255)

    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 0,
        val_split_pct: float = 0.2,
        test_split_pct: float = 0.2,
        **kwargs: Any,
    ) -> None:
        """Initialize a new NASAMarineDebrisDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            val_split_pct: Percentage of the dataset to use as a validation set.
            test_split_pct: Percentage of the dataset to use as a test set.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.NASAMarineDebris`.
        """
        super().__init__(NASAMarineDebris, batch_size, num_workers, **kwargs)

        self.val_split_pct = val_split_pct
        self.test_split_pct = test_split_pct

        self.aug = AugPipe(
            AugmentationSequential(
                K.Normalize(mean=self.mean, std=self.std), data_keys=['image', 'boxes']
            ),
            batch_size,
        )

        self.collate_fn = collate_fn_detection

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        self.dataset = NASAMarineDebris(**self.kwargs)
        generator = torch.Generator().manual_seed(0)
        self.train_dataset, self.val_dataset, self.test_dataset = random_split(
            self.dataset,
            [
                1 - self.val_split_pct - self.test_split_pct,
                self.val_split_pct,
                self.test_split_pct,
            ],
            generator,
        )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""UC Merced dataset."""

import os
from collections.abc import Callable
from typing import ClassVar, cast

import matplotlib.pyplot as plt
import numpy as np
import torchvision.transforms.functional as F
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoClassificationDataset
from .utils import Path, check_integrity, download_url, extract_archive


class UCMerced(NonGeoClassificationDataset):
    """UC Merced Land Use dataset.

    The `UC Merced Land Use <http://weegee.vision.ucmerced.edu/datasets/landuse.html>`_
    dataset is a land use classification dataset of 2.1k 256x256 1ft resolution RGB
    images of urban locations around the U.S. extracted from the USGS National Map Urban
    Area Imagery collection with 21 land use classes (100 images per class).

    Dataset features:

    * land use class labels from around the U.S.
    * three spectral bands - RGB
    * 21 classes

    Dataset classes:

    * agricultural
    * airplane
    * baseballdiamond
    * beach
    * buildings
    * chaparral
    * denseresidential
    * forest
    * freeway
    * golfcourse
    * harbor
    * intersection
    * mediumresidential
    * mobilehomepark
    * overpass
    * parkinglot
    * river
    * runway
    * sparseresidential
    * storagetanks
    * tenniscourt

    This dataset uses the train/val/test splits defined in the "In-domain representation
    learning for remote sensing" paper:

    * https://arxiv.org/abs/1911.06721

    If you use this dataset in your research, please cite the following paper:

    * https://dl.acm.org/doi/10.1145/1869790.1869829
    """

    url = 'https://hf.co/datasets/torchgeo/ucmerced/resolve/d0af6e2eeea2322af86078068bd83337148a2149/UCMerced_LandUse.zip'
    filename = 'UCMerced_LandUse.zip'
    md5 = '5b7ec56793786b6dc8a908e8854ac0e4'

    base_dir = os.path.join('UCMerced_LandUse', 'Images')

    splits = ('train', 'val', 'test')
    split_urls: ClassVar[dict[str, str]] = {
        'train': 'https://storage.googleapis.com/remote_sensing_representations/uc_merced-train.txt',
        'val': 'https://storage.googleapis.com/remote_sensing_representations/uc_merced-val.txt',
        'test': 'https://storage.googleapis.com/remote_sensing_representations/uc_merced-test.txt',
    }
    split_md5s: ClassVar[dict[str, str]] = {
        'train': 'f2fb12eb2210cfb53f93f063a35ff374',
        'val': '11ecabfc52782e5ea6a9c7c0d263aca0',
        'test': '046aff88472d8fc07c4678d03749e28d',
    }

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new UC Merced dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", or "test"
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert split in self.splits
        self.root = root
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self._verify()

        valid_fns = set()
        with open(os.path.join(self.root, f'uc_merced-{split}.txt')) as f:
            for fn in f:
                valid_fns.add(fn.strip())

        def is_in_split(x: Path) -> bool:
            return os.path.basename(x) in valid_fns

        super().__init__(
            root=os.path.join(root, self.base_dir),
            transforms=transforms,
            is_valid_file=is_in_split,
        )

    def _load_image(self, index: int) -> tuple[Tensor, Tensor]:
        """Load a single image and its class label.

        Args:
            index: index to return

        Returns:
            the image and class label
        """
        img, label = super()._load_image(index)
        img = F.resize(img, size=(256, 256), antialias=True)
        return img, label

    def _check_integrity(self) -> bool:
        """Check integrity of dataset.

        Returns:
            True if dataset files are found and/or MD5s match, else False
        """
        integrity: bool = check_integrity(
            os.path.join(self.root, self.filename), self.md5 if self.checksum else None
        )
        return integrity

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        filepath = os.path.join(self.root, self.base_dir)
        if os.path.exists(filepath):
            return

        # Check if zip file already exists (if so then extract)
        if self._check_integrity():
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download and extract the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        download_url(
            self.url,
            self.root,
            filename=self.filename,
            md5=self.md5 if self.checksum else None,
        )
        for split in self.splits:
            download_url(
                self.split_urls[split],
                self.root,
                filename=f'uc_merced-{split}.txt',
                md5=self.split_md5s[split] if self.checksum else None,
            )

    def _extract(self) -> None:
        """Extract the dataset."""
        filepath = os.path.join(self.root, self.filename)
        extract_archive(filepath)

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`NonGeoClassificationDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        image = np.rollaxis(sample['image'].numpy(), 0, 3)

        # Normalize the image if the max value is greater than 1
        if image.max() > 1:
            image = image.astype(np.float32) / 255.0  # Scale to [0, 1]

        label = cast(int, sample['label'].item())
        label_class = self.classes[label]

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = cast(int, sample['prediction'].item())
            prediction_class = self.classes[prediction]

        fig, ax = plt.subplots(figsize=(4, 4))
        ax.imshow(image)
        ax.axis('off')
        if show_titles:
            title = f'Label: {label_class}'
            if showing_predictions:
                title += f'\nPrediction: {prediction_class}'
            ax.set_title(title)

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""FAIR1M datamodule."""

from typing import Any

import torch
from torch import Tensor

from ..datasets import FAIR1M
from .geo import NonGeoDataModule


def collate_fn(batch: list[dict[str, Tensor]]) -> dict[str, Any]:
    """Custom object detection collate fn to handle variable boxes.

    Args:
        batch: list of sample dicts return by dataset

    Returns:
        batch dict output

    .. versionadded:: 0.5
    """
    output: dict[str, Any] = {}
    output['image'] = torch.stack([sample['image'] for sample in batch])

    if 'boxes' in batch[0]:
        output['boxes'] = [sample['boxes'] for sample in batch]
    if 'label' in batch[0]:
        output['label'] = [sample['label'] for sample in batch]

    return output


class FAIR1MDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the FAIR1M dataset.

    .. versionadded:: 0.2
    """

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new FAIR1MDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.FAIR1M`.

        .. versionchanged:: 0.5
           Removed *val_split_pct* and *test_split_pct* parameters.
        """
        super().__init__(FAIR1M, batch_size, num_workers, **kwargs)
        self.collate_fn = collate_fn

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit']:
            self.train_dataset = FAIR1M(split='train', **self.kwargs)
        if stage in ['fit', 'validate']:
            self.val_dataset = FAIR1M(split='val', **self.kwargs)
        if stage in ['predict']:
            # Test set labels are not publicly available
            self.predict_dataset = FAIR1M(split='test', **self.kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""CaBuAr dataset."""

import os
from collections.abc import Callable
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, download_url, lazy_import, percentile_normalization


class CaBuAr(NonGeoDataset):
    """CaBuAr dataset.

    `CaBuAr <https://huggingface.co/datasets/DarthReca/california_burned_areas>`__
    is a dataset for Change detection for Burned area Delineation and part of
    the splits are used for the ChaBuD ECML-PKDD 2023 Discovery Challenge.

    Dataset features:

    * Sentinel-2 multispectral imagery
    * binary masks of burned areas
    * 12 multispectral bands
    * 424 pairs of pre and post images with 20 m per pixel resolution (512x512 px)

    Dataset format:

    * single hdf5 dataset containing images and masks

    Dataset classes:

    0. no change
    1. burned area

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1109/MGRS.2023.3292467

    .. note::

       This dataset requires the following additional library to be installed:

       * `h5py <https://pypi.org/project/h5py/>`_ to load the dataset

    .. versionadded:: 0.6
    """

    all_bands = (
        'B01',
        'B02',
        'B03',
        'B04',
        'B05',
        'B06',
        'B07',
        'B08',
        'B8A',
        'B09',
        'B11',
        'B12',
    )
    rgb_bands = ('B04', 'B03', 'B02')
    folds: ClassVar[dict[str, list[object]]] = {
        'train': [1, 2, 3, 4],
        'val': [0],
        'test': ['chabud'],
    }
    urls = (
        'https://huggingface.co/datasets/DarthReca/california_burned_areas/resolve/main/raw/patched/512x512.hdf5',
        'https://huggingface.co/datasets/DarthReca/california_burned_areas/resolve/main/raw/patched/chabud_test.h5',
    )
    filenames = ('512x512.hdf5', 'chabud_test.h5')
    md5s = ('15d78fb825f9a81dad600db828d22c08', 'a70bb7e4a2788657c2354c4c3d9296fe')

    def __init__(
        self,
        root: Path = 'data',
        split: str = 'train',
        bands: tuple[str, ...] = all_bands,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new CaBuAr dataset instance.

        Args:
            root: root directory where dataset can be found
            split: one of "train", "val", "test"
            bands: the subset of bands to load
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            AssertionError: If ``split`` or ``bands`` arguments are invalid.
            DatasetNotFoundError: If dataset is not found and *download* is False.
            DependencyNotFoundError: If h5py is not installed.
        """
        lazy_import('h5py')

        assert split in self.folds
        assert set(bands) <= set(self.all_bands)

        # Set the file index based on the split
        file_index = 1 if split == 'test' else 0

        self.root = root
        self.split = split
        self.bands = bands
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.filepath = os.path.join(root, self.filenames[file_index])
        self.band_indices = [self.all_bands.index(b) for b in bands]

        self._verify()

        self.uuids = self._load_uuids()

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            sample containing image and mask
        """
        image = self._load_image(index)
        mask = self._load_target(index)

        sample = {'image': image, 'mask': mask}

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.uuids)

    def _load_uuids(self) -> list[str]:
        """Return the image uuids for the given split.

        Returns:
            the image uuids
        """
        h5py = lazy_import('h5py')
        uuids = []
        with h5py.File(self.filepath, 'r') as f:
            for k, v in f.items():
                if v.attrs['fold'] in self.folds[self.split] and 'pre_fire' in v.keys():
                    uuids.append(k)
        return sorted(uuids)

    def _load_image(self, index: int) -> Tensor:
        """Load a single image.

        Args:
            index: index to return

        Returns:
            the image
        """
        h5py = lazy_import('h5py')
        uuid = self.uuids[index]
        with h5py.File(self.filepath, 'r') as f:
            pre_array = f[uuid]['pre_fire'][:]
            post_array = f[uuid]['post_fire'][:]

        # index specified bands and concatenate
        pre_array = pre_array[..., self.band_indices]
        post_array = post_array[..., self.band_indices]
        array = np.concatenate([pre_array, post_array], axis=-1).astype(np.float32)

        tensor = torch.from_numpy(array)
        # Convert from HxWxC to CxHxW
        tensor = tensor.permute((2, 0, 1))
        return tensor

    def _load_target(self, index: int) -> Tensor:
        """Load the target mask for a single image.

        Args:
            index: index to return

        Returns:
            the target mask
        """
        h5py = lazy_import('h5py')
        uuid = self.uuids[index]
        with h5py.File(self.filepath, 'r') as f:
            array = f[uuid]['mask'][:].astype(np.int32).squeeze(axis=-1)

        tensor = torch.from_numpy(array)
        tensor = tensor.to(torch.long)
        return tensor

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        exists = []
        for filename in self.filenames:
            filepath = os.path.join(self.root, filename)
            exists.append(os.path.exists(filepath))

        if all(exists):
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        for url, filename, md5 in zip(self.urls, self.filenames, self.md5s):
            filepath = os.path.join(self.root, filename)
            if not os.path.exists(filepath):
                download_url(
                    url,
                    self.root,
                    filename=filename,
                    md5=md5 if self.checksum else None,
                )

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional suptitle to use for figure

        Returns:
            a matplotlib Figure with the rendered sample
        """
        rgb_indices = []
        for band in self.rgb_bands:
            if band in self.bands:
                rgb_indices.append(self.bands.index(band))
            else:
                raise ValueError("Dataset doesn't contain some of the RGB bands")

        mask = sample['mask'].numpy()
        image_pre = sample['image'][: len(self.bands)][rgb_indices].numpy()
        image_post = sample['image'][len(self.bands) :][rgb_indices].numpy()
        image_pre = percentile_normalization(image_pre)
        image_post = percentile_normalization(image_post)

        ncols = 3

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            prediction = sample['prediction']
            ncols += 1

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(10, ncols * 5))

        axs[0].imshow(np.transpose(image_pre, (1, 2, 0)))
        axs[0].axis('off')
        axs[1].imshow(np.transpose(image_post, (1, 2, 0)))
        axs[1].axis('off')
        axs[2].imshow(mask)
        axs[2].axis('off')

        if showing_predictions:
            axs[3].imshow(prediction)
            axs[3].axis('off')

        if show_titles:
            axs[0].set_title('Image Pre')
            axs[1].set_title('Image Post')
            axs[2].set_title('Mask')
            if showing_predictions:
                axs[3].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Northeastern China Crop Map Dataset."""

from collections.abc import Callable, Iterable
from typing import Any, ClassVar

import matplotlib.pyplot as plt
import torch
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import RasterDataset
from .utils import BoundingBox, Path, download_url


class NCCM(RasterDataset):
    """The Northeastern China Crop Map Dataset.

    Link: https://www.nature.com/articles/s41597-021-00827-9

    This dataset produced annual 10-m crop maps of the
    major crops (maize, soybean, and rice)
    in Northeast China from 2017 to 2019, using hierarchial mapping strategies,
    random forest classifiers, interpolated and
    smoothed 10-day Sentinel-2 time series data and
    optimized features from spectral, temporal and
    textural characteristics of the land surface.
    The resultant maps have high overall accuracies (OA)
    based on ground truth data. The dataset contains information
    specific to three years: 2017, 2018, 2019.

    The dataset contains 5 classes:

    0. paddy rice
    1. maize
    2. soybean
    3. others crops and lands
    4. nodata

    Dataset format:

    * Three .TIF files containing the labels
    * JavaScript code to download images from the dataset.

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1038/s41597-021-00827-9

    .. versionadded:: 0.6
    """

    filename_regex = r'CDL(?P<date>\d{4})_clip'
    filename_glob = 'CDL*.*'

    date_format = '%Y'
    is_image = False
    urls: ClassVar[dict[int, str]] = {
        2019: 'https://figshare.com/ndownloader/files/25070540',
        2018: 'https://figshare.com/ndownloader/files/25070624',
        2017: 'https://figshare.com/ndownloader/files/25070582',
    }
    md5s: ClassVar[dict[int, str]] = {
        2019: '0d062bbd42e483fdc8239d22dba7020f',
        2018: 'b3bb4894478d10786aa798fb11693ec1',
        2017: 'd047fbe4a85341fa6248fd7e0badab6c',
    }
    fnames: ClassVar[dict[int, str]] = {
        2019: 'CDL2019_clip.tif',
        2018: 'CDL2018_clip1.tif',
        2017: 'CDL2017_clip.tif',
    }

    cmap: ClassVar[dict[int, tuple[int, int, int, int]]] = {
        0: (0, 255, 0, 255),
        1: (255, 0, 0, 255),
        2: (255, 255, 0, 255),
        3: (128, 128, 128, 255),
        15: (255, 255, 255, 255),
    }

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | None = None,
        years: list[int] = [2019],
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new dataset.

        Args:
            paths: one or more root directories to search or files to load
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS
                (defaults to the resolution of the first file found)
            years: list of years for which to use nccm layers
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 after downloading files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert set(years) <= self.md5s.keys(), (
            'NCCM data product only exists for the following years: '
            f'{list(self.md5s.keys())}.'
        )
        self.paths = paths
        self.years = years
        self.download = download
        self.checksum = checksum
        self.ordinal_map = torch.full((max(self.cmap.keys()) + 1,), 4, dtype=self.dtype)
        self.ordinal_cmap = torch.zeros((5, 4), dtype=torch.uint8)

        self._verify()
        super().__init__(paths, crs, res, transforms=transforms, cache=cache)

        for i, (k, v) in enumerate(self.cmap.items()):
            self.ordinal_map[k] = i
            self.ordinal_cmap[i] = torch.tensor(v)

    def __getitem__(self, query: BoundingBox) -> dict[str, Any]:
        """Retrieve mask and metadata indexed by query.

        Args:
            query: (minx, maxx, miny, maxy, mint, maxt) coordinates to index

        Returns:
            sample of mask and metadata at that index

        Raises:
            IndexError: if query is not found in the index
        """
        sample = super().__getitem__(query)
        sample['mask'] = self.ordinal_map[sample['mask']]
        return sample

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        if self.files:
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()

    def _download(self) -> None:
        """Download the dataset."""
        for year in self.years:
            download_url(
                self.urls[year],
                self.paths,
                filename=self.fnames[year],
                md5=self.md5s[year] if self.checksum else None,
            )

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`NCCM.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        mask = sample['mask'].squeeze()
        ncols = 1

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'].squeeze()
            ncols = 2

        fig, axs = plt.subplots(
            nrows=1, ncols=ncols, figsize=(ncols * 4, 4), squeeze=False
        )

        axs[0, 0].imshow(self.ordinal_cmap[mask], interpolation='none')
        axs[0, 0].axis('off')

        if show_titles:
            axs[0, 0].set_title('Mask')

        if showing_predictions:
            axs[0, 1].imshow(self.ordinal_cmap[pred], interpolation='none')
            axs[0, 1].axis('off')
            if show_titles:
                axs[0, 1].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Self-Supervised Learning for Earth Observation Benchmark Datasets."""

import glob
import os
from collections.abc import Callable
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
import rasterio
import torch
from matplotlib.figure import Figure
from torch import Tensor

from .cdl import CDL
from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .nlcd import NLCD
from .utils import Path, download_url, extract_archive


class SSL4EOLBenchmark(NonGeoDataset):
    """SSL4EO Landsat Benchmark Evaluation Dataset.

    Dataset is intended to be used for evaluation of SSL techniques. Each
    benchmark dataset consists of 25,000 images with corresponding land
    cover classification masks.

    Dataset format:

    * Input landsat image and single channel mask
    * 25,000 total samples split into train, val, test (70%, 15%, 15%)
    * NLCD dataset version has 17 classes
    * CDL dataset version has 134 classes

    Each patch has the following properties:

    * 264 x 264 pixels
    * Resampled to 30 m resolution (7920 x 7920 m)
    * Single multispectral GeoTIFF file

    If you use this dataset in your research, please cite the following paper:

    * https://proceedings.neurips.cc/paper_files/paper/2023/hash/bbf7ee04e2aefec136ecf60e346c2e61-Abstract-Datasets_and_Benchmarks.html

    .. versionadded:: 0.5
    """

    url = 'https://hf.co/datasets/torchgeo/ssl4eo-l-benchmark/resolve/da96ae2b04cb509710b72fce9131c2a3d5c211c2/{}.tar.gz'

    valid_sensors = ('tm_toa', 'etm_toa', 'etm_sr', 'oli_tirs_toa', 'oli_sr')
    valid_products = ('cdl', 'nlcd')
    valid_splits = ('train', 'val', 'test')

    image_root = 'ssl4eo_l_{}_benchmark'
    img_md5s: ClassVar[dict[str, str]] = {
        'tm_toa': '8e3c5bcd56d3780a442f1332013b8d15',
        'etm_toa': '1b051c7fe4d61c581b341370c9e76f1f',
        'etm_sr': '34a24fa89a801654f8d01e054662c8cd',
        'oli_tirs_toa': '6e9d7cf0392e1de2cbdb39962ba591aa',
        'oli_sr': '0700cd15cc2366fe68c2f8c02fa09a15',
    }

    mask_dir_dict: ClassVar[dict[str, str]] = {
        'tm_toa': 'ssl4eo_l_tm_{}',
        'etm_toa': 'ssl4eo_l_etm_{}',
        'etm_sr': 'ssl4eo_l_etm_{}',
        'oli_tirs_toa': 'ssl4eo_l_oli_{}',
        'oli_sr': 'ssl4eo_l_oli_{}',
    }
    mask_md5s: ClassVar[dict[str, dict[str, str]]] = {
        'tm': {
            'cdl': '3d676770ffb56c7e222a7192a652a846',
            'nlcd': '261149d7614fcfdcb3be368eefa825c7',
        },
        'etm': {
            'cdl': '008098c968544049eaf7b307e14241de',
            'nlcd': '9c031049d665202ba42ac1d89b687999',
        },
        'oli': {
            'cdl': '1cb057de6eafeca975deb35cb9fb036f',
            'nlcd': '9de0d6d4d0b94313b80450f650813922',
        },
    }

    year_dict: ClassVar[dict[str, int]] = {
        'tm_toa': 2011,
        'etm_toa': 2019,
        'etm_sr': 2019,
        'oli_tirs_toa': 2019,
        'oli_sr': 2019,
    }

    rgb_indices: ClassVar[dict[str, list[int]]] = {
        'tm_toa': [2, 1, 0],
        'etm_toa': [2, 1, 0],
        'etm_sr': [2, 1, 0],
        'oli_tirs_toa': [3, 2, 1],
        'oli_sr': [3, 2, 1],
    }

    split_percentages = (0.7, 0.15, 0.15)

    cmaps: ClassVar[dict[str, dict[int, tuple[int, int, int, int]]]] = {
        'nlcd': NLCD.cmap,
        'cdl': CDL.cmap,
    }

    def __init__(
        self,
        root: Path = 'data',
        sensor: str = 'oli_sr',
        product: str = 'cdl',
        split: str = 'train',
        classes: list[int] | None = None,
        transforms: Callable[[dict[str, Tensor]], dict[str, Tensor]] | None = None,
        download: bool = False,
        checksum: bool = False,
    ) -> None:
        """Initialize a new SSL4EO Landsat Benchmark instance.

        Args:
            root: root directory where dataset can be found
            sensor: one of ['etm_toa', 'etm_sr', 'oli_tirs_toa, 'oli_sr']
            product: mask target, one of ['cdl', 'nlcd']
            split: dataset split, one of ['train', 'val', 'test']
            classes: list of classes to include, the rest will be mapped to 0
                (defaults to all classes for the chosen product)
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory
            checksum: if True, check the MD5 after downloading files (may be slow)

        Raises:
            AssertionError: if any arguments are invalid
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert (
            sensor in self.valid_sensors
        ), f'Only supports one of {self.valid_sensors}, but found {sensor}.'
        self.sensor = sensor
        assert (
            product in self.valid_products
        ), f'Only supports one of {self.valid_products}, but found {product}.'
        self.product = product
        assert (
            split in self.valid_splits
        ), f'Only supports one of {self.valid_splits}, but found {split}.'
        self.split = split

        self.cmap = self.cmaps[product]
        if classes is None:
            classes = list(self.cmap.keys())

        assert (
            set(classes) <= self.cmap.keys()
        ), f'Only the following classes are valid: {list(self.cmap.keys())}.'
        assert 0 in classes, 'Classes must include the background class: 0'

        self.root = root
        self.classes = classes
        self.transforms = transforms
        self.download = download
        self.checksum = checksum
        self.ordinal_map = torch.zeros(max(self.cmap.keys()) + 1, dtype=torch.long)
        self.ordinal_cmap = torch.zeros((len(self.classes), 4), dtype=torch.uint8)
        self.img_dir_name = self.image_root.format(self.sensor)
        self.mask_dir_name = self.mask_dir_dict[self.sensor].format(self.product)

        self._verify()

        self.sample_collection = self.retrieve_sample_collection()

        # train, val, test split
        np.random.seed(0)
        sizes = (np.array(self.split_percentages) * len(self.sample_collection)).astype(
            int
        )
        cutoffs = np.cumsum(sizes)[:-1]
        sample_indices = np.arange(len(self.sample_collection))
        np.random.shuffle(sample_indices)
        groups = np.split(sample_indices, cutoffs)
        split_indices = {'train': groups[0], 'val': groups[1], 'test': groups[2]}[
            self.split
        ]

        self.sample_collection = [self.sample_collection[idx] for idx in split_indices]

        # Map chosen classes to ordinal numbers, all others mapped to background class
        for v, k in enumerate(self.classes):
            self.ordinal_map[k] = v
            self.ordinal_cmap[v] = torch.tensor(self.cmap[k])

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted files already exist
        img_pathname = os.path.join(self.root, self.img_dir_name, '**', 'all_bands.tif')
        exists = []
        exists.append(bool(glob.glob(img_pathname, recursive=True)))
        mask_pathname = os.path.join(
            self.root,
            self.mask_dir_name,
            '**',
            f'{self.product}_{self.year_dict[self.sensor]}.tif',
        )
        exists.append(bool(glob.glob(mask_pathname, recursive=True)))

        if all(exists):
            return
        # Check if the tar.gz files have already been downloaded
        exists = []
        img_pathname = os.path.join(self.root, f'{self.img_dir_name}.tar.gz')
        exists.append(os.path.exists(img_pathname))

        mask_pathname = os.path.join(self.root, f'{self.mask_dir_name}.tar.gz')
        exists.append(os.path.exists(mask_pathname))

        if all(exists):
            self._extract()
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self._extract()

    def _download(self) -> None:
        """Download the dataset."""
        # download imagery
        download_url(
            self.url.format(self.img_dir_name),
            self.root,
            md5=self.img_md5s[self.sensor] if self.checksum else None,
        )
        # download mask
        download_url(
            self.url.format(self.mask_dir_name),
            self.root,
            md5=(
                self.mask_md5s[self.sensor.split('_')[0]][self.product]
                if self.checksum
                else None
            ),
        )

    def _extract(self) -> None:
        """Extract the dataset."""
        img_pathname = os.path.join(self.root, f'{self.img_dir_name}.tar.gz')
        extract_archive(img_pathname)

        mask_pathname = os.path.join(self.root, f'{self.mask_dir_name}.tar.gz')
        extract_archive(mask_pathname)

    def __getitem__(self, index: int) -> dict[str, Tensor]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            image and sample
        """
        img_path, mask_path = self.sample_collection[index]

        sample = {
            'image': self._load_image(img_path),
            'mask': self._load_mask(mask_path),
        }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.sample_collection)

    def retrieve_sample_collection(self) -> list[tuple[str, str]]:
        """Retrieve paths to samples in data directory."""
        img_paths = glob.glob(
            os.path.join(self.root, self.img_dir_name, '**', 'all_bands.tif'),
            recursive=True,
        )
        img_paths = sorted(img_paths)
        sample_collection: list[tuple[str, str]] = []
        for img_path in img_paths:
            mask_path = img_path.replace(self.img_dir_name, self.mask_dir_name).replace(
                'all_bands.tif', f'{self.product}_{self.year_dict[self.sensor]}.tif'
            )
            sample_collection.append((img_path, mask_path))
        return sample_collection

    def _load_image(self, path: Path) -> Tensor:
        """Load the input image.

        Args:
            path: path to input image

        Returns:
            image
        """
        with rasterio.open(path) as src:
            image = torch.from_numpy(src.read()).float()
        return image

    def _load_mask(self, path: Path) -> Tensor:
        """Load the mask.

        Args:
            path: path to mask

        Retuns:
            mask
        """
        with rasterio.open(path) as src:
            mask = torch.from_numpy(src.read()).long()
        mask = self.ordinal_map[mask]
        return mask

    def plot(
        self,
        sample: dict[str, Tensor],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        ncols = 2
        image = sample['image'][self.rgb_indices[self.sensor]].permute(1, 2, 0)
        image = image / 255

        mask = sample['mask'].squeeze(0)

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'].squeeze(0)
            ncols = 3

        fig, ax = plt.subplots(ncols=ncols, figsize=(4 * ncols, 4))
        ax[0].imshow(image)
        ax[0].axis('off')
        ax[1].imshow(self.ordinal_cmap[mask], interpolation='none')
        ax[1].axis('off')
        if show_titles:
            ax[0].set_title('Image')
            ax[1].set_title('Mask')

        if showing_predictions:
            ax[2].imshow(self.ordinal_cmap[pred], interpolation='none')
            if show_titles:
                ax[2].set_title('Prediction')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""ChaBuD datamodule."""

from typing import Any

import torch
from einops import repeat

from ..datasets import ChaBuD
from .geo import NonGeoDataModule


class ChaBuDDataModule(NonGeoDataModule):
    """LightningDataModule implementation for the ChaBuD dataset.

    Uses the train/val splits from the dataset

    .. versionadded:: 0.6
    """

    # min/max values computed on train set using 2/98 percentiles
    min = torch.tensor(
        [0.0, 1.0, 73.0, 39.0, 46.0, 25.0, 26.0, 21.0, 17.0, 1.0, 20.0, 21.0]
    )
    max = torch.tensor(
        [
            1926.0,
            2174.0,
            2527.0,
            2950.0,
            3237.0,
            3717.0,
            4087.0,
            4271.0,
            4290.0,
            4219.0,
            4568.0,
            3753.0,
        ]
    )

    def __init__(
        self, batch_size: int = 64, num_workers: int = 0, **kwargs: Any
    ) -> None:
        """Initialize a new ChaBuDDataModule instance.

        Args:
            batch_size: Size of each mini-batch.
            num_workers: Number of workers for parallel data loading.
            **kwargs: Additional keyword arguments passed to
                :class:`~torchgeo.datasets.ChaBuD`.
        """
        bands = kwargs.get('bands', ChaBuD.all_bands)
        band_indices = [ChaBuD.all_bands.index(b) for b in bands]
        mins = self.min[band_indices]
        maxs = self.max[band_indices]

        # Change detection, 2 images from different times
        mins = repeat(mins, 'c -> (t c)', t=2)
        maxs = repeat(maxs, 'c -> (t c)', t=2)

        self.mean = mins
        self.std = maxs - mins

        super().__init__(ChaBuD, batch_size, num_workers, **kwargs)

    def setup(self, stage: str) -> None:
        """Set up datasets.

        Args:
            stage: Either 'fit', 'validate', 'test', or 'predict'.
        """
        if stage in ['fit', 'validate']:
            self.train_dataset = ChaBuD(split='train', **self.kwargs)
            self.val_dataset = ChaBuD(split='val', **self.kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Western USA Live Fuel Moisture Dataset."""

import glob
import json
import os
from collections.abc import Callable, Iterable
from typing import Any

import pandas as pd
import torch

from .errors import DatasetNotFoundError
from .geo import NonGeoDataset
from .utils import Path, which


class WesternUSALiveFuelMoisture(NonGeoDataset):
    """Western USA Live Fuel Moisture Dataset.

    This tabular style dataset contains fuel moisture
    (mass of water in vegetation) and remotely sensed variables
    in the western United States. It contains 2615 datapoints and 138
    variables. For more details see the
    `dataset page <https://beta.source.coop/stanford/sar-moisture-conent/>`_.

    Dataset Format:

    * .geojson file for each datapoint

    Dataset Features:

    * 138 remote sensing derived variables, some with a time dependency
    * 2615 datapoints with regression target of predicting fuel moisture

    If you use this dataset in your research, please cite the following paper:

    * https://doi.org/10.1016/j.rse.2020.111797

    .. note::

       This dataset requires the following additional library to be installed:

       * `azcopy <https://github.com/Azure/azure-storage-azcopy>`_: to download the
         dataset from Source Cooperative.

    .. versionadded:: 0.5
    """

    url = 'https://radiantearth.blob.core.windows.net/mlhub/su-sar-moisture-content'

    label_name = 'percent(t)'

    all_variable_names = (
        # "date",
        'slope(t)',
        'elevation(t)',
        'canopy_height(t)',
        'forest_cover(t)',
        'silt(t)',
        'sand(t)',
        'clay(t)',
        'vv(t)',
        'vh(t)',
        'red(t)',
        'green(t)',
        'blue(t)',
        'swir(t)',
        'nir(t)',
        'ndvi(t)',
        'ndwi(t)',
        'nirv(t)',
        'vv_red(t)',
        'vv_green(t)',
        'vv_blue(t)',
        'vv_swir(t)',
        'vv_nir(t)',
        'vv_ndvi(t)',
        'vv_ndwi(t)',
        'vv_nirv(t)',
        'vh_red(t)',
        'vh_green(t)',
        'vh_blue(t)',
        'vh_swir(t)',
        'vh_nir(t)',
        'vh_ndvi(t)',
        'vh_ndwi(t)',
        'vh_nirv(t)',
        'vh_vv(t)',
        'slope(t-1)',
        'elevation(t-1)',
        'canopy_height(t-1)',
        'forest_cover(t-1)',
        'silt(t-1)',
        'sand(t-1)',
        'clay(t-1)',
        'vv(t-1)',
        'vh(t-1)',
        'red(t-1)',
        'green(t-1)',
        'blue(t-1)',
        'swir(t-1)',
        'nir(t-1)',
        'ndvi(t-1)',
        'ndwi(t-1)',
        'nirv(t-1)',
        'vv_red(t-1)',
        'vv_green(t-1)',
        'vv_blue(t-1)',
        'vv_swir(t-1)',
        'vv_nir(t-1)',
        'vv_ndvi(t-1)',
        'vv_ndwi(t-1)',
        'vv_nirv(t-1)',
        'vh_red(t-1)',
        'vh_green(t-1)',
        'vh_blue(t-1)',
        'vh_swir(t-1)',
        'vh_nir(t-1)',
        'vh_ndvi(t-1)',
        'vh_ndwi(t-1)',
        'vh_nirv(t-1)',
        'vh_vv(t-1)',
        'slope(t-2)',
        'elevation(t-2)',
        'canopy_height(t-2)',
        'forest_cover(t-2)',
        'silt(t-2)',
        'sand(t-2)',
        'clay(t-2)',
        'vv(t-2)',
        'vh(t-2)',
        'red(t-2)',
        'green(t-2)',
        'blue(t-2)',
        'swir(t-2)',
        'nir(t-2)',
        'ndvi(t-2)',
        'ndwi(t-2)',
        'nirv(t-2)',
        'vv_red(t-2)',
        'vv_green(t-2)',
        'vv_blue(t-2)',
        'vv_swir(t-2)',
        'vv_nir(t-2)',
        'vv_ndvi(t-2)',
        'vv_ndwi(t-2)',
        'vv_nirv(t-2)',
        'vh_red(t-2)',
        'vh_green(t-2)',
        'vh_blue(t-2)',
        'vh_swir(t-2)',
        'vh_nir(t-2)',
        'vh_ndvi(t-2)',
        'vh_ndwi(t-2)',
        'vh_nirv(t-2)',
        'vh_vv(t-2)',
        'slope(t-3)',
        'elevation(t-3)',
        'canopy_height(t-3)',
        'forest_cover(t-3)',
        'silt(t-3)',
        'sand(t-3)',
        'clay(t-3)',
        'vv(t-3)',
        'vh(t-3)',
        'red(t-3)',
        'green(t-3)',
        'blue(t-3)',
        'swir(t-3)',
        'nir(t-3)',
        'ndvi(t-3)',
        'ndwi(t-3)',
        'nirv(t-3)',
        'vv_red(t-3)',
        'vv_green(t-3)',
        'vv_blue(t-3)',
        'vv_swir(t-3)',
        'vv_nir(t-3)',
        'vv_ndvi(t-3)',
        'vv_ndwi(t-3)',
        'vv_nirv(t-3)',
        'vh_red(t-3)',
        'vh_green(t-3)',
        'vh_blue(t-3)',
        'vh_swir(t-3)',
        'vh_nir(t-3)',
        'vh_ndvi(t-3)',
        'vh_ndwi(t-3)',
        'vh_nirv(t-3)',
        'vh_vv(t-3)',
        'lat',
        'lon',
    )

    def __init__(
        self,
        root: Path = 'data',
        input_features: Iterable[str] = all_variable_names,
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        download: bool = False,
    ) -> None:
        """Initialize a new Western USA Live Fuel Moisture Dataset.

        Args:
            root: root directory where dataset can be found
            input_features: which input features to include
            transforms: a function/transform that takes input sample and its target as
                entry and returns a transformed version
            download: if True, download dataset and store it in the root directory

        Raises:
            AssertionError: if ``input_features`` contains invalid variable names
            DatasetNotFoundError: If dataset is not found and *download* is False.
        """
        assert set(input_features) <= set(self.all_variable_names)

        self.root = root
        self.input_features = input_features
        self.transforms = transforms
        self.download = download

        self._verify()

        self.dataframe = self._load_data()

    def __len__(self) -> int:
        """Return the number of data points in the dataset.

        Returns:
            length of the dataset
        """
        return len(self.dataframe)

    def __getitem__(self, index: int) -> dict[str, Any]:
        """Return an index within the dataset.

        Args:
            index: index to return

        Returns:
            input features and target at that index
        """
        data = self.dataframe.iloc[index, :]

        sample = {
            'input': torch.tensor(
                data.drop([self.label_name]).values, dtype=torch.float32
            ),
            'label': torch.tensor(data[self.label_name], dtype=torch.float32),
        }

        if self.transforms is not None:
            sample = self.transforms(sample)

        return sample

    def _load_data(self) -> pd.DataFrame:
        """Load data from individual files into pandas dataframe.

        Returns:
            the features and label
        """
        data_rows = []
        for path in sorted(self.files):
            with open(path) as f:
                content = json.load(f)
                data_dict = content['properties']
                data_dict['lon'] = content['geometry']['coordinates'][0]
                data_dict['lat'] = content['geometry']['coordinates'][1]
                data_rows.append(data_dict)

        df = pd.DataFrame(data_rows)
        df = df[[*self.input_features, self.label_name]]
        return df

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the files already exist
        file_glob = os.path.join(self.root, '**', 'feature_*.geojson')
        self.files = glob.glob(file_glob, recursive=True)
        if self.files:
            return

        # Check if the user requested to download the dataset
        if not self.download:
            raise DatasetNotFoundError(self)

        # Download the dataset
        self._download()
        self.files = glob.glob(file_glob, recursive=True)

    def _download(self) -> None:
        """Download the dataset and extract it."""
        os.makedirs(self.root, exist_ok=True)
        azcopy = which('azcopy')
        azcopy('sync', self.url, self.root, '--recursive=true')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""European Digital Elevation Model (EU-DEM) dataset."""

import glob
import os
from collections.abc import Callable, Iterable
from typing import Any, ClassVar

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from rasterio.crs import CRS

from .errors import DatasetNotFoundError
from .geo import RasterDataset
from .utils import Path, check_integrity, extract_archive


class EUDEM(RasterDataset):
    """European Digital Elevation Model (EU-DEM) Dataset.

    `EU-DEM <https://www.eea.europa.eu/en/datahub/datahubitem-view/d08852bc-7b5f-4835-a776-08362e2fbf4b>`__
    is a Digital Elevation Model of reference for the entire European region.

    Dataset features:

    * DEMs at 25 m per pixel spatial resolution (~40,000x40,0000 px)
    * vertical accuracy of +/- 7 m RMSE
    * data fused from `ASTER GDEM
      <https://lpdaac.usgs.gov/news/nasa-and-meti-release-aster-global-dem-version-3/>`_,
      `SRTM <https://science.jpl.nasa.gov/projects/srtm/>`_ and Russian topomaps

    Dataset format:

    * DEMs are single-channel tif files

    .. versionadded:: 0.3
    """

    is_image = False
    filename_glob = 'eu_dem_v11_*.TIF'
    zipfile_glob = 'eu_dem_v11_*[A-Z0-9].zip'
    filename_regex = '(?P<name>[eudem_v11]{10})_(?P<id>[A-Z0-9]{6})'

    md5s: ClassVar[dict[str, str]] = {
        'eu_dem_v11_E00N20.zip': '96edc7e11bc299b994e848050d6be591',
        'eu_dem_v11_E10N00.zip': 'e14be147ac83eddf655f4833d55c1571',
        'eu_dem_v11_E10N10.zip': '2eb5187e4d827245b33768404529c709',
        'eu_dem_v11_E10N20.zip': '1afc162eb131841aed0d00b692b870a8',
        'eu_dem_v11_E20N10.zip': '77b040791b9fb7de271b3f47130b4e0c',
        'eu_dem_v11_E20N20.zip': '89b965abdcb1dbd479c61117f55230c8',
        'eu_dem_v11_E20N30.zip': 'f5cb1b05813ae8ffc9e70f0ad56cc372',
        'eu_dem_v11_E20N40.zip': '81be551ff646802d7d820385de7476e9',
        'eu_dem_v11_E20N50.zip': 'bbc351713ea3eb7e9eb6794acb9e4bc8',
        'eu_dem_v11_E30N10.zip': '68fb95aac33a025c4f35571f32f237ff',
        'eu_dem_v11_E30N20.zip': 'da8ad029f9cc1ec9234ea3e7629fe18d',
        'eu_dem_v11_E30N30.zip': 'de27c78d0176e45aec5c9e462a95749c',
        'eu_dem_v11_E30N40.zip': '4c00e58b624adfc4a5748c922e77ee40',
        'eu_dem_v11_E30N50.zip': '4a21a88f4d2047b8995d1101df0b3a77',
        'eu_dem_v11_E40N10.zip': '32fdf4572581eddc305a21c5d2f4bc81',
        'eu_dem_v11_E40N20.zip': '71b027f29258493dd751cfd63f08578f',
        'eu_dem_v11_E40N30.zip': 'c6c21289882c1f74fc4649d255302c64',
        'eu_dem_v11_E40N40.zip': '9f26e6e47f4160ef8ea5200e8cf90a45',
        'eu_dem_v11_E40N50.zip': 'a8c3c1c026cdd1537b8a3822c15834d9',
        'eu_dem_v11_E50N10.zip': '9584273c7708b8e935f2bac3e30c19c6',
        'eu_dem_v11_E50N20.zip': '8efdea43e7b6819861935d5a768a55f2',
        'eu_dem_v11_E50N30.zip': 'e39e58df1c13ac35eb0b29fb651f313c',
        'eu_dem_v11_E50N40.zip': 'd84395ab52ad254d930db17398fffc50',
        'eu_dem_v11_E50N50.zip': '6abe852f4a20962db0e355ffc0d695a4',
        'eu_dem_v11_E60N10.zip': 'b6a3b8a39a4efc01c7e2cd8418672559',
        'eu_dem_v11_E60N20.zip': '71dc3c55ab5c90628ce2149dbd60f090',
        'eu_dem_v11_E70N20.zip': '5342465ad60cf7d28a586c9585179c35',
    }

    def __init__(
        self,
        paths: Path | Iterable[Path] = 'data',
        crs: CRS | None = None,
        res: float | None = None,
        transforms: Callable[[dict[str, Any]], dict[str, Any]] | None = None,
        cache: bool = True,
        checksum: bool = False,
    ) -> None:
        """Initialize a new Dataset instance.

        Args:
            paths: one or more root directories to search or files to load, here
                the collection of individual zip files for each tile should be found
            crs: :term:`coordinate reference system (CRS)` to warp to
                (defaults to the CRS of the first file found)
            res: resolution of the dataset in units of CRS
                (defaults to the resolution of the first file found)
            transforms: a function/transform that takes an input sample
                and returns a transformed version
            cache: if True, cache file handle to speed up repeated sampling
            checksum: if True, check the MD5 of the downloaded files (may be slow)

        Raises:
            DatasetNotFoundError: If dataset is not found.

        .. versionchanged:: 0.5
           *root* was renamed to *paths*.
        """
        self.paths = paths
        self.checksum = checksum

        self._verify()

        super().__init__(paths, crs, res, transforms=transforms, cache=cache)

    def _verify(self) -> None:
        """Verify the integrity of the dataset."""
        # Check if the extracted file already exists
        if self.files:
            return

        # Check if the zip files have already been downloaded
        assert isinstance(self.paths, str | os.PathLike)
        pathname = os.path.join(self.paths, self.zipfile_glob)
        if glob.glob(pathname):
            for zipfile in glob.iglob(pathname):
                filename = os.path.basename(zipfile)
                if self.checksum and not check_integrity(zipfile, self.md5s[filename]):
                    raise RuntimeError('Dataset found, but corrupted.')
                extract_archive(zipfile)
            return

        raise DatasetNotFoundError(self)

    def plot(
        self,
        sample: dict[str, Any],
        show_titles: bool = True,
        suptitle: str | None = None,
    ) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`RasterDataset.__getitem__`
            show_titles: flag indicating whether to show titles above each panel
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        mask = sample['mask'].squeeze()
        ncols = 1

        showing_predictions = 'prediction' in sample
        if showing_predictions:
            pred = sample['prediction'].squeeze()
            ncols = 2

        fig, axs = plt.subplots(nrows=1, ncols=ncols, figsize=(ncols * 4, 4))

        if showing_predictions:
            axs[0].imshow(mask)
            axs[0].axis('off')
            axs[1].imshow(pred)
            axs[1].axis('off')
            if show_titles:
                axs[0].set_title('Mask')
                axs[1].set_title('Prediction')
        else:
            axs.imshow(mask)
            axs.axis('off')
            if show_titles:
                axs.set_title('Mask')

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Common datamodule utilities."""

import math
from collections.abc import Callable, Iterable
from typing import Any

import numpy as np
import torch
from einops import rearrange
from torch import Tensor
from torch.nn import Module


# Based on lightning_lite.utilities.exceptions
class MisconfigurationException(Exception):
    """Exception used to inform users of misuse with Lightning."""


class AugPipe(Module):
    """Pipeline for applying augmentations sequentially on select data keys.

    .. versionadded:: 0.6
    """

    def __init__(
        self, augs: Callable[[dict[str, Any]], dict[str, Any]], batch_size: int
    ) -> None:
        """Initialize a new AugPipe instance.

        Args:
            augs: Augmentations to apply.
            batch_size: Batch size
        """
        super().__init__()
        self.augs = augs
        self.batch_size = batch_size

    def forward(self, batch: dict[str, Tensor]) -> dict[str, Tensor]:
        """Apply the augmentation.

        Args:
            batch: Input batch.

        Returns:
            Augmented batch.
        """
        batch_len = len(batch['image'])
        for bs in range(batch_len):
            batch_dict = {
                'image': batch['image'][bs],
                'labels': batch['labels'][bs],
                'boxes': batch['boxes'][bs],
            }

            if 'masks' in batch:
                batch_dict['masks'] = batch['masks'][bs]

            batch_dict = self.augs(batch_dict)

            batch['image'][bs] = batch_dict['image']
            batch['labels'][bs] = batch_dict['labels']
            batch['boxes'][bs] = batch_dict['boxes']

            if 'masks' in batch:
                batch['masks'][bs] = batch_dict['masks']

        # Stack images
        batch['image'] = rearrange(batch['image'], 'b () c h w -> b c h w')

        return batch


def collate_fn_detection(batch: list[dict[str, Tensor]]) -> dict[str, Any]:
    """Custom collate fn for object detection and instance segmentation.

    Args:
        batch: list of sample dicts return by dataset

    Returns:
        batch dict output

    .. versionadded:: 0.6
    """
    output: dict[str, Any] = {}
    output['image'] = [sample['image'] for sample in batch]
    output['boxes'] = [sample['boxes'].float() for sample in batch]
    if 'labels' in batch[0]:
        output['labels'] = [sample['labels'] for sample in batch]
    else:
        output['labels'] = [
            torch.tensor([1] * len(sample['boxes'])) for sample in batch
        ]

    if 'masks' in batch[0]:
        output['masks'] = [sample['masks'] for sample in batch]
    return output


def group_shuffle_split(
    groups: Iterable[Any],
    train_size: float | None = None,
    test_size: float | None = None,
    random_state: int | None = None,
) -> tuple[list[int], list[int]]:
    """Method for performing a single group-wise shuffle split of data.

    Loosely based off of :class:`sklearn.model_selection.GroupShuffleSplit`.

    Args:
        groups: a sequence of group values used to split. Should be in the same order as
            the data you want to split.
        train_size: the proportion of groups to include in the train split. If None,
            then it is set to complement `test_size`.
        test_size: the proportion of groups to include in the test split (rounded up).
            If None, then it is set to complement `train_size`.
        random_state: controls the random splits (passed a seed to a
            numpy.random.Generator), set for reproducible splits.

    Returns:
        train_indices, test_indices

    Raises:
        ValueError if `train_size` and `test_size` do not sum to 1, aren't in the range
            (0,1), or are both None.
        ValueError if the number of training or testing groups turns out to be 0.
    """
    if train_size is None and test_size is None:
        raise ValueError('You must specify `train_size`, `test_size`, or both.')
    if (train_size is not None and test_size is not None) and (
        not math.isclose(train_size + test_size, 1)
    ):
        raise ValueError('`train_size` and `test_size` must sum to 1.')

    if train_size is None and test_size is not None:
        train_size = 1 - test_size
    if test_size is None and train_size is not None:
        test_size = 1 - train_size

    assert train_size is not None and test_size is not None

    if train_size <= 0 or train_size >= 1 or test_size <= 0 or test_size >= 1:
        raise ValueError('`train_size` and `test_size` must be in the range (0,1).')

    group_vals = sorted(set(groups))
    n_groups = len(group_vals)
    n_test_groups = round(n_groups * test_size)
    n_train_groups = n_groups - n_test_groups

    if n_train_groups == 0 or n_test_groups == 0:
        raise ValueError(
            f'{n_groups} groups were found, however the current settings of '
            + '`train_size` and `test_size` result in 0 training or testing groups.'
        )

    generator = np.random.default_rng(seed=random_state)
    train_group_vals = set(
        generator.choice(group_vals, size=n_train_groups, replace=False)
    )

    train_idxs = []
    test_idxs = []
    for i, group_val in enumerate(groups):
        if group_val in train_group_vals:
            train_idxs.append(i)
        else:
            test_idxs.append(i)

    return train_idxs, test_idxs
